from contextlib import AbstractAsyncContextManager
from functools import partial
from typing import (
    AsyncIterator,
    Iterator,
    List,
    Optional,
)

from pydantic import StrictStr

from .generated.openapi_aclient import ApiClient as AApiClient
from .generated.openapi_aclient.api.v1_api import V1Api as AModelsApi
from .generated.openapi_aclient.models.model_list import ModelList as AModelItem
from .generated.openapi_aclient.models.model_request import (
    ModelRequest as AModelRequest,
)
from .generated.openapi_aclient.models.paginated_model_list_list import (
    PaginatedModelListList as APaginatedModelListList,
)
from .generated.openapi_client import ApiClient
from .generated.openapi_client.api.v1_api import V1Api as ModelsApi
from .generated.openapi_client.models.model_list import ModelList as ModelItem
from .generated.openapi_client.models.model_request import ModelRequest
from .utils import (
    ClientContextCallable,
    iterate_cursor_list,
    with_async_client,
    with_sync_client,
)


class Models:
    """Models (sub) API

    Note:

      The construction of the API instance should be handled by
      accesing an attribute of a :class:`root.client.RootSignals` instance.
    """

    def __init__(
        self,
        client_context: ClientContextCallable,
    ):
        self.client_context = client_context

    @with_sync_client
    def list(
        self,
        *,
        capable_of: Optional[List[str]] = None,
        limit: int = 100,
        _client: ApiClient,
    ) -> Iterator[ModelItem]:
        """Iterate through the models.

        Args:
          limit: Number of entries to iterate through at most.
          capable_of: List of capabilities to filter the models by.
        """

        api_instance = ModelsApi(_client)
        yield from iterate_cursor_list(partial(api_instance.v1_models_list, capable_of=capable_of), limit=limit)

    async def alist(
        self,
        *,
        capable_of: Optional[List[str]] = None,
        limit: int = 100,
    ) -> AsyncIterator[AModelItem]:
        """Asynchronously iterate through the models.

        Args:
          limit: Number of entries to iterate through at most.
          capable_of: List of capabilities to filter the models by.
        """
        context = self.client_context()
        assert isinstance(context, AbstractAsyncContextManager), "This method is not available in synchronous mode"

        async with context as client:
            api_instance = AModelsApi(client)
            partial_list = partial(api_instance.v1_models_list, capable_of=capable_of)

            cursor: Optional[StrictStr] = None
            while limit > 0:
                result: APaginatedModelListList = await partial_list(page_size=limit, cursor=cursor)
                if not result.results:
                    return
                used_results = result.results[:limit]
                limit -= len(used_results)
                for used_result in used_results:
                    yield used_result

                if not (cursor := result.next):
                    return

    @with_sync_client
    def create(
        self,
        *,
        name: str,
        model: Optional[str] = None,
        default_key: Optional[str] = None,
        max_output_token_count: Optional[int] = None,
        max_token_count: Optional[int] = None,
        url: Optional[str] = None,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> str:
        """Create a new model and return its ID.

        Args:
          name: The unique identifier for the model instance  (e.g. "google/gemma-2-9b").
          model: The base model name to be used. Defaults to name.
          default_key: The default API key required for the model, if applicable.
          max_output_token_count: The maximum number of tokens to output.
          max_token_count: The maximum number of tokens to process.
          url: Optional URL pointing to the model's endpoint.
        """
        request = ModelRequest(
            name=name,
            model=model,
            default_key=default_key,
            max_output_token_count=max_output_token_count,
            max_token_count=max_token_count,
            url=url,
        )

        api_instance = ModelsApi(_client)
        return api_instance.v1_models_create(model_request=request, _request_timeout=_request_timeout).id

    @with_async_client
    async def acreate(
        self,
        *,
        name: str,
        model: Optional[str] = None,
        default_key: Optional[str] = None,
        max_output_token_count: Optional[int] = None,
        max_token_count: Optional[int] = None,
        url: Optional[str] = None,
        _request_timeout: Optional[int] = None,
        _client: AApiClient,
    ) -> str:
        """Asynchronously create a new model and return its ID.

        Args:
          name: The unique identifier for the model instance  (e.g. "google/gemma-2-9b").
          model: The base model name to be used. Defaults to name.
          default_key: The default API key required for the model, if applicable.
          max_output_token_count: The maximum number of tokens to output.
          max_token_count: The maximum number of tokens to process.
          url: Optional URL pointing to the model's endpoint.
        """

        request = AModelRequest(
            name=name,
            model=model,
            default_key=default_key,
            max_output_token_count=max_output_token_count,
            max_token_count=max_token_count,
            url=url,
        )
        api_instance = AModelsApi(_client)
        created_model = await api_instance.v1_models_create(model_request=request, _request_timeout=_request_timeout)
        return created_model.id

    @with_sync_client
    def delete(
        self,
        model_id: str,
        *,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> None:
        """
        Delete the model.
        """
        api_instance = ModelsApi(_client)
        return api_instance.v1_models_destroy(id=model_id, _request_timeout=_request_timeout)

    @with_async_client
    async def adelete(
        self,
        model_id: str,
        *,
        _request_timeout: Optional[int] = None,
        _client: AApiClient,
    ) -> None:
        """
        Asynchronously delete the model.
        """
        api_instance = AModelsApi(_client)
        return await api_instance.v1_models_destroy(id=model_id, _request_timeout=_request_timeout)

    # TODO: update



from __future__ import annotations

from contextlib import AbstractAsyncContextManager
from functools import partial
from typing import TYPE_CHECKING, AsyncIterator, Iterator, List, Optional, cast

from pydantic import StrictStr

from .generated.openapi_aclient import ApiClient as AApiClient
from .generated.openapi_aclient.api.v1_api import V1Api as AObjectivesApi
from .generated.openapi_aclient.models.evaluator_execution_functions_request import (
    EvaluatorExecutionFunctionsRequest as AEvaluatorExecutionFunctionsRequest,
)
from .generated.openapi_aclient.models.objective import Objective as AOpenApiObjective
from .generated.openapi_aclient.models.objective_execution_request import (
    ObjectiveExecutionRequest as AObjectiveExecutionRequest,
)
from .generated.openapi_aclient.models.objective_list import ObjectiveList as AObjectiveList
from .generated.openapi_aclient.models.objective_request import ObjectiveRequest as AObjectiveRequest
from .generated.openapi_aclient.models.paginated_objective_list import (
    PaginatedObjectiveList as APaginatedObjectiveList,
)
from .generated.openapi_aclient.models.paginated_objective_list_list import (
    PaginatedObjectiveListList as APaginatedObjectiveListList,
)
from .generated.openapi_aclient.models.patched_objective_request import (
    PatchedObjectiveRequest as APatchedObjectiveRequest,
)
from .generated.openapi_aclient.models.validator_execution_result import (
    ValidatorExecutionResult as AValidatorExecutionResult,
)
from .generated.openapi_client import ApiClient
from .generated.openapi_client.api.v1_api import V1Api as ObjectivesApi
from .generated.openapi_client.models.evaluator_execution_functions_request import (
    EvaluatorExecutionFunctionsRequest,
)
from .generated.openapi_client.models.objective import Objective as OpenApiObjective
from .generated.openapi_client.models.objective_execution_request import ObjectiveExecutionRequest
from .generated.openapi_client.models.objective_list import ObjectiveList
from .generated.openapi_client.models.objective_request import ObjectiveRequest
from .generated.openapi_client.models.paginated_objective_list import PaginatedObjectiveList
from .generated.openapi_client.models.patched_objective_request import PatchedObjectiveRequest
from .generated.openapi_client.models.validator_execution_result import ValidatorExecutionResult
from .skills import Skills
from .utils import ClientContextCallable, iterate_cursor_list, with_async_client, with_sync_client
from .validators import AValidator

if TYPE_CHECKING:
    from .validators import Validator


class Versions:
    """Version listing (sub)API

    Note that this should not be directly instantiated.
    """

    def __init__(self, client_context: ClientContextCallable):
        self.client_context = client_context

    @with_sync_client
    def list(self, objective_id: str, *, _client: ApiClient) -> PaginatedObjectiveList:
        """List all versions of an objective.

        Args:
          objective_id: The objective to list the versions for
        """

        api_instance = ObjectivesApi(_client)
        return api_instance.get_a_list_of_all_versions_of_an_objective(id=objective_id)

    async def alist(self, objective_id: str) -> APaginatedObjectiveList:
        """Asynchronously list all versions of an objective.

        Args:
          objective_id: The objective to list the versions for
        """

        context = self.client_context()
        assert isinstance(context, AbstractAsyncContextManager), "This method is not available in synchronous mode"
        async with context as client:
            api_instance = AObjectivesApi(client)
            return await api_instance.get_a_list_of_all_versions_of_an_objective(id=objective_id)


class Objective(OpenApiObjective):
    """Wrapper for a single Objective.

    For available attributes, please check the (automatically
    generated) superclass documentation.
    """

    client_context: ClientContextCallable

    @classmethod
    def _wrap(cls, apiobj: OpenApiObjective, client_context: ClientContextCallable) -> "Objective":
        if not isinstance(apiobj, OpenApiObjective):
            raise ValueError(f"Wrong instance in _wrap: {apiobj!r}")
        obj = cast(Objective, apiobj)
        obj.__class__ = cls
        obj.client_context = client_context
        return obj

    @with_sync_client
    def run(
        self,
        *,
        response: str,
        request: Optional[str] = None,
        contexts: Optional[List[str]] = None,
        functions: Optional[List[EvaluatorExecutionFunctionsRequest]] = None,
        expected_output: Optional[str] = None,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> ValidatorExecutionResult:
        """
        Run all validators associated with the objective.

        Args:
          response: LLM output.
          request: The prompt sent to the LLM. Optional.
          contexts: Optional documents passed to RAG evaluators
          functions: Optional function definitions to LLM tool call validation
          expected_output: Optional expected output for the evaluators
        """

        api_instance = ObjectivesApi(_client)
        skill_execution_request = ObjectiveExecutionRequest(
            request=request,
            response=response,
            contexts=contexts,
            functions=functions,
            expected_output=expected_output,
        )
        return api_instance.v1_objectives_objectives_execute_create(
            objective_id=self.id,
            objective_execution_request=skill_execution_request,
            _request_timeout=_request_timeout,
        )


class AObjective(AOpenApiObjective):
    """
    Wrapper for a single Objective.

    For available attributes, please check the (automatically
    generated) superclass documentation.
    """

    client_context: ClientContextCallable

    @classmethod
    async def _awrap(cls, apiobj: AOpenApiObjective, client_context: ClientContextCallable) -> "AObjective":
        if not isinstance(apiobj, AOpenApiObjective):
            raise ValueError(f"Wrong instance in _wrap: {apiobj!r}")
        obj = cast(AObjective, apiobj)
        obj.__class__ = cls
        obj.client_context = client_context
        return obj

    @with_async_client
    async def arun(
        self,
        *,
        response: str,
        request: Optional[str] = None,
        contexts: Optional[List[str]] = None,
        functions: Optional[List[AEvaluatorExecutionFunctionsRequest]] = None,
        expected_output: Optional[str] = None,
        _request_timeout: Optional[int] = None,
        _client: AApiClient,
    ) -> AValidatorExecutionResult:
        """
        Asynchronously run all validators associated with the objective.

        Args:
          response: LLM output.
          request: The prompt sent to the LLM. Optional.
          contexts: Optional documents passed to RAG evaluators
          functions: Optional function definitions to LLM tool call validation
          expected_output: Optional expected output for the evaluators
        """

        api_instance = AObjectivesApi(_client)
        skill_execution_request = AObjectiveExecutionRequest(
            request=request,
            response=response,
            contexts=contexts,
            functions=functions,
            expected_output=expected_output,
        )
        return await api_instance.v1_objectives_objectives_execute_create(
            objective_id=self.id,
            objective_execution_request=skill_execution_request,
            _request_timeout=_request_timeout,
        )


class Objectives:
    """
    Objectives API

    Note:

      The construction of the API instance should be handled by
      accesing an attribute of a :class:`root.client.RootSignals` instance.
    """

    def __init__(self, client_context: ClientContextCallable):
        self.client_context = client_context
        self.versions = Versions(client_context)

    @with_sync_client
    def create(
        self,
        *,
        intent: Optional[str] = None,
        validators: Optional[List[Validator]] = None,
        test_dataset_id: Optional[str] = None,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> Objective:
        """
        Create a new objective and return its ID.

        Args:
          intent: The intent of the objective.
          validators: An optional list of validators.
          test_dataset_id: The ID of the test dataset
        """

        skills = Skills(self.client_context)
        request = ObjectiveRequest(
            intent=intent,
            validators=[validator._to_request(skills) for validator in validators or []],
            test_dataset_id=test_dataset_id,
        )
        api_instance = ObjectivesApi(_client)
        objective = api_instance.v1_objectives_create(objective_request=request)
        return self.get(objective.id, _request_timeout=_request_timeout)

    @with_async_client
    async def acreate(
        self,
        *,
        intent: Optional[str] = None,
        validators: Optional[List[AValidator]] = None,
        test_dataset_id: Optional[str] = None,
        _request_timeout: Optional[int] = None,
        _client: AApiClient,
    ) -> AObjective:
        """
        Asynchronously create a new objective and return its ID.

        Args:
          intent: The intent of the objective.
          validators: An optional list of validators.
          test_dataset_id: The ID of the test dataset
        """

        skills = Skills(self.client_context)
        request = AObjectiveRequest(
            intent=intent,
            validators=[await avalidator._ato_request(skills) for avalidator in validators or []],
            test_dataset_id=test_dataset_id,
        )
        api_instance = AObjectivesApi(_client)
        objective = await api_instance.v1_objectives_create(objective_request=request)
        return await self.aget(objective.id, _request_timeout=_request_timeout)

    @with_sync_client
    def get(
        self,
        objective_id: str,
        *,
        _client: ApiClient,
        _request_timeout: Optional[int] = None,
    ) -> Objective:
        """
        Get an objective by ID.

        Args:
          objective_id: The objective to be fetched.
        """

        api_instance = ObjectivesApi(_client)
        return Objective._wrap(
            api_instance.v1_objectives_retrieve(id=objective_id, _request_timeout=_request_timeout),
            client_context=self.client_context,
        )

    @with_async_client
    async def aget(
        self,
        objective_id: str,
        *,
        _client: AApiClient,
        _request_timeout: Optional[int] = None,
    ) -> AObjective:
        """
        Asynchronously get an objective by ID.

        Args:
          objective_id: The objective to be fetched.
        """

        api_instance = AObjectivesApi(_client)
        return await AObjective._awrap(
            await api_instance.v1_objectives_retrieve(id=objective_id, _request_timeout=_request_timeout),
            client_context=self.client_context,
        )

    @with_sync_client
    def delete(self, objective_id: str, *, _request_timeout: Optional[int] = None, _client: ApiClient) -> None:
        """
        Delete the objective from the registry.

        Args:
          objective_id: The objective to be deleted.
        """

        api_instance = ObjectivesApi(_client)
        return api_instance.v1_objectives_destroy(id=objective_id, _request_timeout=_request_timeout)

    @with_async_client
    async def adelete(self, objective_id: str, *, _request_timeout: Optional[int] = None, _client: AApiClient) -> None:
        """
        Asynchronously delete the objective from the registry.

        Args:
          objective_id: The objective to be deleted.
        """

        api_instance = AObjectivesApi(_client)
        return await api_instance.v1_objectives_destroy(id=objective_id, _request_timeout=_request_timeout)

    @with_sync_client
    def list(self, *, intent: Optional[str] = None, limit: int = 100, _client: ApiClient) -> Iterator[ObjectiveList]:
        """
        Iterate through the objectives.

        Args:
          intent: Specific intent the returned objectives must match.
          limit: Number of entries to iterate through at most.
        """

        api_instance = ObjectivesApi(_client)
        yield from iterate_cursor_list(partial(api_instance.v1_objectives_list, intent=intent), limit=limit)

    @with_async_client
    async def alist(self, *, intent: Optional[str] = None, limit: int = 100) -> AsyncIterator[AObjectiveList]:
        """
        Asynchronously iterate through the objectives.

        Args:
          intent: Specific intent the returned objectives must match.
          limit: Number of entries to iterate through at most.

        """

        context = self.client_context()
        assert isinstance(context, AbstractAsyncContextManager), "This method is not available in synchronous mode"
        async with context as client:
            api_instance = AObjectivesApi(client)
            partial_list = partial(api_instance.v1_objectives_list, intent=intent)

            cursor: Optional[StrictStr] = None
            while limit > 0:
                result: APaginatedObjectiveListList = await partial_list(page_size=limit, cursor=cursor)
                if not result.results:
                    return

                used_results = result.results[:limit]
                limit -= len(used_results)
                for used_result in used_results:
                    yield used_result

                if not (cursor := result.next):
                    return

    @with_sync_client
    def run(
        self,
        objective_id: str,
        *,
        response: str,
        request: Optional[str] = None,
        contexts: Optional[List[str]] = None,
        functions: Optional[List[EvaluatorExecutionFunctionsRequest]] = None,
        expected_output: Optional[str] = None,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> ValidatorExecutionResult:
        """
        Run all validators associated with an objective.

        Args:
          response: LLM output.
          request: The prompt sent to the LLM. Optional.
          contexts: Optional documents passed to RAG evaluators
        """

        api_instance = ObjectivesApi(_client)
        skill_execution_request = ObjectiveExecutionRequest(
            request=request,
            response=response,
            contexts=contexts,
            functions=functions,
            expected_output=expected_output,
        )
        return api_instance.v1_objectives_objectives_execute_create(
            objective_id=objective_id,
            objective_execution_request=skill_execution_request,
            _request_timeout=_request_timeout,
        )

    @with_async_client
    async def arun(
        self,
        objective_id: str,
        *,
        _client: AApiClient,
        response: str,
        request: Optional[str] = None,
        contexts: Optional[List[str]] = None,
        functions: Optional[List[AEvaluatorExecutionFunctionsRequest]] = None,
        expected_output: Optional[str] = None,
        _request_timeout: Optional[int] = None,
    ) -> AValidatorExecutionResult:
        """
        Asynchronously run all validators associated with an objective.

        Args:
          response: LLM output.
          request: The prompt sent to the LLM. Optional.
          contexts: Optional documents passed to RAG evaluators
        """

        api_instance = AObjectivesApi(_client)
        skill_execution_request = AObjectiveExecutionRequest(
            request=request,
            response=response,
            contexts=contexts,
            functions=functions,
            expected_output=expected_output,
        )
        return await api_instance.v1_objectives_objectives_execute_create(
            objective_id=objective_id,
            objective_execution_request=skill_execution_request,
            _request_timeout=_request_timeout,
        )

    @with_sync_client
    def update(
        self,
        objective_id: str,
        *,
        intent: Optional[str] = None,
        validators: Optional[List[Validator]] = None,
        test_dataset_id: Optional[str] = None,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> Objective:
        """
        Update an existing objective.

        Args:
          objective_id: The objective to be updated.
          intent: The intent of the objective.
          validators: An optional list of validators.
        """

        skills = Skills(self.client_context)
        request = PatchedObjectiveRequest(
            intent=intent,
            validators=[validator._to_request(skills) for validator in validators] if validators else None,
            test_dataset_id=test_dataset_id,
        )
        api_instance = ObjectivesApi(_client)
        return Objective._wrap(
            api_instance.v1_objectives_partial_update(
                id=objective_id,
                patched_objective_request=request,
                _request_timeout=_request_timeout,
            ),
            client_context=self.client_context,
        )

    @with_async_client
    async def aupdate(
        self,
        objective_id: str,
        *,
        intent: Optional[str] = None,
        validators: Optional[List[AValidator]] = None,
        test_dataset_id: Optional[str] = None,
        _request_timeout: Optional[int] = None,
        _client: AApiClient,
    ) -> AObjective:
        """
        Asynchronously update an existing objective.

        Args:
          objective_id: The objective to be updated.
          intent: The intent of the objective.
          validators: An optional list of validators.
        """

        skills = Skills(self.client_context)
        request = APatchedObjectiveRequest(
            intent=intent,
            validators=[await validator._ato_request(skills) for validator in validators] if validators else None,
            test_dataset_id=test_dataset_id,
        )
        api_instance = AObjectivesApi(_client)
        return await AObjective._awrap(
            await api_instance.v1_objectives_partial_update(
                id=objective_id,
                patched_objective_request=request,
                _request_timeout=_request_timeout,
            ),
            client_context=self.client_context,
        )



from __future__ import annotations

from contextlib import AbstractAsyncContextManager
from functools import partial
from typing import AsyncIterator, Iterator, List, Optional, Union, cast

from pydantic import StrictStr

from .generated.openapi_aclient import ApiClient as AApiClient
from .generated.openapi_aclient.api.beta_api import BetaApi as ABetaApi
from .generated.openapi_aclient.models.evaluator_execution_functions_request import (
    EvaluatorExecutionFunctionsRequest as AEvaluatorExecutionFunctionsRequest,
)
from .generated.openapi_aclient.models.evaluator_reference_request import (
    EvaluatorReferenceRequest as AEvaluatorReferenceRequest,
)
from .generated.openapi_aclient.models.judge import Judge as AOpenApiJudge
from .generated.openapi_aclient.models.judge_execution_request import (
    JudgeExecutionRequest as AJudgeExecutionRequest,
)
from .generated.openapi_aclient.models.judge_execution_response import (
    JudgeExecutionResponse as AJudgeExecutionResponse,
)
from .generated.openapi_aclient.models.judge_list import JudgeList as AJudgeList
from .generated.openapi_aclient.models.paginated_judge_list_list import (
    PaginatedJudgeListList as APaginatedJudgeListList,
)
from .generated.openapi_aclient.models.patched_judge_request import (
    PatchedJudgeRequest as APatchedJudgeRequest,
)
from .generated.openapi_client import ApiClient
from .generated.openapi_client.api.beta_api import BetaApi
from .generated.openapi_client.models.evaluator_execution_functions_request import (
    EvaluatorExecutionFunctionsRequest,
)
from .generated.openapi_client.models.evaluator_reference_request import EvaluatorReferenceRequest
from .generated.openapi_client.models.judge import Judge as OpenApiJudge
from .generated.openapi_client.models.judge_execution_request import JudgeExecutionRequest
from .generated.openapi_client.models.judge_execution_response import JudgeExecutionResponse
from .generated.openapi_client.models.judge_list import JudgeList
from .generated.openapi_client.models.paginated_judge_list_list import PaginatedJudgeListList
from .generated.openapi_client.models.patched_judge_request import PatchedJudgeRequest
from .utils import ClientContextCallable, with_async_client, with_sync_client


class Judge(OpenApiJudge):
    """Wrapper for a single Judge.

    For available attributes, please check the (automatically
    generated) superclass documentation.
    """

    client_context: ClientContextCallable

    @classmethod
    def _wrap(cls, apiobj: Union[OpenApiJudge, JudgeList], client_context: ClientContextCallable) -> Judge:
        """Wrap API object into a Judge instance."""
        if not isinstance(apiobj, (OpenApiJudge, JudgeList)):
            raise ValueError(f"Wrong instance in _wrap: {apiobj!r}")
        obj = cast(Judge, apiobj)
        obj.__class__ = cls
        obj.client_context = client_context
        return obj

    @with_sync_client
    def run(
        self,
        *,
        response: str,
        request: Optional[str] = None,
        contexts: Optional[List[str]] = None,
        functions: Optional[List[EvaluatorExecutionFunctionsRequest]] = None,
        expected_output: Optional[str] = None,
        tags: Optional[List[str]] = None,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> JudgeExecutionResponse:
        """
        Run the judge.

        Args:
          response: LLM output to evaluate
          request: The prompt sent to the LLM. Optional.
          contexts: Optional documents passed to RAG evaluators
          functions: Optional functions to execute
          expected_output: Optional expected output
          tags: Optional tags to add to the judge execution
          _request_timeout: Optional timeout for the request
        """
        api_instance = BetaApi(_client)
        execution_request = JudgeExecutionRequest(
            request=request,
            response=response,
            contexts=contexts,
            functions=functions,
            expected_output=expected_output,
            tags=tags,
        )
        return api_instance.beta_judges_execute_create(
            judge_id=self.id,
            judge_execution_request=execution_request,
            _request_timeout=_request_timeout,
        )


class AJudge(AOpenApiJudge):
    """
    Async wrapper for a single Judge.

    For available attributes, please check the (automatically
    generated) superclass documentation.
    """

    client_context: ClientContextCallable

    @classmethod
    async def _awrap(cls, apiobj: Union[AOpenApiJudge, AJudgeList], client_context: ClientContextCallable) -> AJudge:
        if not isinstance(apiobj, (AOpenApiJudge, AJudgeList)):
            raise ValueError(f"Wrong instance in _wrap: {apiobj!r}")
        obj = cast(AJudge, apiobj)
        obj.__class__ = cls
        obj.client_context = client_context
        return obj

    @with_async_client
    async def arun(
        self,
        *,
        response: str,
        request: Optional[str] = None,
        contexts: Optional[List[str]] = None,
        functions: Optional[List[AEvaluatorExecutionFunctionsRequest]] = None,
        expected_output: Optional[str] = None,
        tags: Optional[List[str]] = None,
        _request_timeout: Optional[int] = None,
        _client: AApiClient,
    ) -> AJudgeExecutionResponse:
        """
        Asynchronously run the judge.

        Args:
          response: LLM output to evaluate
          request: The prompt sent to the LLM. Optional.
          contexts: Optional documents passed to RAG evaluators
          functions: Optional functions to execute
          expected_output: Optional expected output
          tags: Optional tags to add to the judge execution
          _request_timeout: Optional timeout for the request
        """
        api_instance = ABetaApi(_client)
        execution_request = AJudgeExecutionRequest(
            contexts=contexts,
            functions=functions,
            expected_output=expected_output,
            request=request,
            response=response,
            tags=tags,
        )
        return await api_instance.beta_judges_execute_create(
            judge_id=self.id,
            judge_execution_request=execution_request,
            _request_timeout=_request_timeout,
        )


class Judges:
    """
    Judges API

    Note:
        The construction of the API instance should be handled by
        accessing an attribute of a :class:`root.client.RootSignals` instance.
    """

    def __init__(self, client_context: ClientContextCallable):
        self.client_context = client_context

    @with_sync_client
    def get(self, judge_id: str, *, _request_timeout: Optional[int] = None, _client: ApiClient) -> Judge:
        """
        Get a judge by ID.

        Args:
          judge_id: The judge to be fetched.
        """
        api_instance = BetaApi(_client)
        return Judge._wrap(
            api_instance.beta_judges_retrieve(id=judge_id, _request_timeout=_request_timeout),
            client_context=self.client_context,
        )

    @with_async_client
    async def aget(self, judge_id: str, *, _request_timeout: Optional[int] = None, _client: AApiClient) -> AJudge:
        """
        Asynchronously get a judge by ID.

        Args:
          judge_id: The judge to be fetched.
        """
        api_instance = ABetaApi(_client)
        return await AJudge._awrap(
            await api_instance.beta_judges_retrieve(id=judge_id, _request_timeout=_request_timeout),
            client_context=self.client_context,
        )

    @with_sync_client
    def delete(self, judge_id: str, *, _request_timeout: Optional[int] = None, _client: ApiClient) -> None:
        """
        Delete the judge.

        Args:
          judge_id: The judge to be deleted.
        """
        api_instance = BetaApi(_client)
        return api_instance.beta_judges_destroy(id=judge_id, _request_timeout=_request_timeout)

    @with_async_client
    async def adelete(self, judge_id: str, *, _request_timeout: Optional[int] = None, _client: AApiClient) -> None:
        """
        Asynchronously delete the judge.

        Args:
          judge_id: The judge to be deleted.
        """
        api_instance = ABetaApi(_client)
        return await api_instance.beta_judges_destroy(id=judge_id, _request_timeout=_request_timeout)

    @with_sync_client
    def list(self, *, limit: int = 100, _client: ApiClient) -> Iterator[Judge]:
        """
        Iterate through the judges.

        Args:
          limit: Number of entries to iterate through at most.
        """
        api_instance = BetaApi(_client)
        cursor: Optional[StrictStr] = None
        while limit > 0:
            result: PaginatedJudgeListList = api_instance.beta_judges_list(page_size=limit, cursor=cursor)
            if not result.results:
                return

            used_results = result.results[:limit]
            limit -= len(used_results)
            for judge in used_results:
                yield Judge._wrap(judge, client_context=self.client_context)

            if not (cursor := result.next):
                return

    async def alist(self, *, limit: int = 100) -> AsyncIterator[AJudge]:
        """
        Asynchronously iterate through the judges.

        Args:
          limit: Number of entries to iterate through at most.
        """
        context = self.client_context()
        assert isinstance(context, AbstractAsyncContextManager), "This method is not available in synchronous mode"
        async with context as client:
            api_instance = ABetaApi(client)
            partial_list = partial(api_instance.beta_judges_list)

            cursor: Optional[StrictStr] = None
            while limit > 0:
                result: APaginatedJudgeListList = await partial_list(page_size=limit, cursor=cursor)
                if not result.results:
                    return

                used_results = result.results[:limit]
                limit -= len(used_results)
                for judge in used_results:
                    yield await AJudge._awrap(judge, client_context=self.client_context)

                if not (cursor := result.next):
                    return

    @with_sync_client
    def update(
        self,
        judge_id: str,
        *,
        name: Optional[str] = None,
        evaluator_references: Optional[List[EvaluatorReferenceRequest]] = None,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> Judge:
        """
        Update an existing judge.

        Args:
          judge_id: The judge to be updated.
          name: New name for the judge
          evaluator_references: New list of evaluator references
        """
        api_instance = BetaApi(_client)
        request = PatchedJudgeRequest(
            name=name,
            evaluator_references=evaluator_references,
        )
        return Judge._wrap(
            api_instance.beta_judges_partial_update(
                id=judge_id,
                patched_judge_request=request,
                _request_timeout=_request_timeout,
            ),
            client_context=self.client_context,
        )

    @with_async_client
    async def aupdate(
        self,
        judge_id: str,
        *,
        name: Optional[str] = None,
        evaluator_references: Optional[List[AEvaluatorReferenceRequest]] = None,
        _request_timeout: Optional[int] = None,
        _client: AApiClient,
    ) -> AJudge:
        """
        Asynchronously update an existing judge.

        Args:
          judge_id: The judge to be updated.
          name: New name for the judge
          evaluator_references: New list of evaluator references
        """
        api_instance = ABetaApi(_client)
        request = APatchedJudgeRequest(
            name=name,
            evaluator_references=evaluator_references,
        )
        return await AJudge._awrap(
            await api_instance.beta_judges_partial_update(
                id=judge_id,
                patched_judge_request=request,
                _request_timeout=_request_timeout,
            ),
            client_context=self.client_context,
        )

    @with_sync_client
    def run(
        self,
        judge_id: str,
        *,
        response: str,
        request: Optional[str] = None,
        contexts: Optional[List[str]] = None,
        functions: Optional[List[EvaluatorExecutionFunctionsRequest]] = None,
        expected_output: Optional[str] = None,
        tags: Optional[List[str]] = None,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> JudgeExecutionResponse:
        """
        Run a judge directly by ID.

        Args:
          judge_id: ID of the judge to run
          response: LLM output to evaluate
          request: The prompt sent to the LLM. Optional.
          contexts: Optional documents passed to RAG evaluators
          functions: Optional functions to execute
          expected_output: Optional expected output
          tags: Optional tags to add to the judge execution
          _request_timeout: Optional timeout for the request
        """
        api_instance = BetaApi(_client)
        execution_request = JudgeExecutionRequest(
            request=request,
            response=response,
            contexts=contexts,
            functions=functions,
            expected_output=expected_output,
            tags=tags,
        )
        return api_instance.beta_judges_execute_create(
            judge_id=judge_id,
            judge_execution_request=execution_request,
            _request_timeout=_request_timeout,
        )

    @with_async_client
    async def arun(
        self,
        judge_id: str,
        *,
        response: str,
        request: Optional[str] = None,
        contexts: Optional[List[str]] = None,
        functions: Optional[List[AEvaluatorExecutionFunctionsRequest]] = None,
        expected_output: Optional[str] = None,
        tags: Optional[List[str]] = None,
        _request_timeout: Optional[int] = None,
        _client: AApiClient,
    ) -> AJudgeExecutionResponse:
        """
        Asynchronously run a judge directly by ID.

        Args:
          judge_id: ID of the judge to run
          response: LLM output to evaluate
          request: The prompt sent to the LLM. Optional.
          contexts: Optional documents passed to RAG evaluators
          functions: Optional functions to execute
          expected_output: Optional expected output
          tags: Optional tags to add to the judge execution
          _request_timeout: Optional timeout for the request
        """
        api_instance = ABetaApi(_client)
        execution_request = AJudgeExecutionRequest(
            contexts=contexts,
            functions=functions,
            expected_output=expected_output,
            request=request,
            response=response,
            tags=tags,
        )
        return await api_instance.beta_judges_execute_create(
            judge_id=judge_id,
            judge_execution_request=execution_request,
            _request_timeout=_request_timeout,
        )



from root import RootSignals

client = RootSignals()

network_troubleshooting_evaluator = client.evaluators.create(
    name="Network Troubleshooting",
    predicate="""Assess the response for technical accuracy and appropriateness in the context of network troubleshooting.
            Is the advice technically sound and relevant to the user's question?
            Does the troubleshooting process effectively address the likely causes of the issue?
            Is the proposed solution valid and safe to implement?

            User question: {{request}}

            Chatbot response: {{response}}
            """,
    intent="To measure the technical accuracy and appropriateness of network troubleshooting responses",
    model="gemini-2.0-flash",  # Check client.models.list() for all available models. You can also add your own model.
)

response = network_troubleshooting_evaluator.run(
    request="My internet is not working.",
    response="""
    I'm sorry to hear that your internet isn't working.
    Let's troubleshoot this step by step.
    """,
)

print(response.score)
print(response.justification)



from root import RootSignals

aclient = RootSignals(run_async=True)


async def main():
    custom_evaluator = await aclient.evaluators.acreate(
        name="Direct language",
        predicate="Is the following text clear and has no weasel words: {{response}}",
        intent="Is the language direct and unambiguous",
        model="gpt-4o",
    )

    response = await custom_evaluator.arun(response="It will probably rain tomorrow.")

    print(response.score)



from unittest.mock import patch

import pytest

from root.client import RootSignals
from root.generated.openapi_client.models.evaluator_calibration_output import EvaluatorCalibrationOutput
from root.generated.openapi_client.models.evaluator_calibration_result import EvaluatorCalibrationResult
from root.skills import CalibrateBatchParameters


@pytest.fixture
def mock_skills_calibrate_api():
    with patch("root.generated.openapi_client.api.v1_api.V1Api.v1_skills_calibrate_create") as mock:
        yield mock


def test_calibrate_evaluator_batch(mock_skills_calibrate_api):
    mock_skills_calibrate_api.return_value = [
        EvaluatorCalibrationOutput(
            result=EvaluatorCalibrationResult(
                score=0.8,
                expected_score=0.75,
                llm_output="output",
                model="gpt-4",
                rendered_prompt="What is the weather today?",
                cost=0.1,
                execution_log_id="1",
            ),
            row_number=1,
            variables={},
        ),
        EvaluatorCalibrationOutput(
            result=EvaluatorCalibrationResult(
                score=0.9,
                expected_score=0.55,
                llm_output="output",
                model="gpt-4",
                rendered_prompt="What is the weather today?",
                cost=0.1,
                execution_log_id="2",
            ),
            row_number=2,
            variables={},
        ),
    ]

    client = RootSignals(api_key="fake")

    params = [
        CalibrateBatchParameters(
            name="With gpt-4",
            prompt="What is the weather today?",
            model="gpt-4",
            pii_filter=False,
            reference_variables=None,
            input_variables=None,
            data_loaders=None,
        ),
        CalibrateBatchParameters(
            name="With gpt-4-turbo",
            prompt="What is the weather today?",
            model="gpt-4-turbo",
            pii_filter=False,
            reference_variables=None,
            input_variables=None,
            data_loaders=None,
        ),
    ]

    result = client.evaluators.calibrate_batch(
        evaluator_definitions=params, test_data=[["0.4", "LLM output"], ["0.6", "LLM output 2"]]
    )

    assert result.rms_errors_model == {
        "gpt-4": pytest.approx(0.2499, rel=1e-3),
        "gpt-4-turbo": pytest.approx(0.2499, rel=1e-3),
    }

    assert result.mae_errors_model == {
        "gpt-4": pytest.approx(0.2, rel=1e-3),
        "gpt-4-turbo": pytest.approx(0.2, rel=1e-3),
    }

    assert result.rms_errors_prompt == {
        "What is the weather today?": pytest.approx(0.25, rel=1e-3),
    }
    assert len(result.results) == 4



from concurrent.futures import ThreadPoolExecutor

from root import RootSignals

# Connect to the Root Signals API
client = RootSignals()


def main():
    response = "This is polite and clear."

    tasks = [
        (client.evaluators.Politeness, response),
        (client.evaluators.Clarity, response),
    ]

    with ThreadPoolExecutor() as executor:
        future_to_eval = {func.__name__: executor.submit(func, response) for func, response in tasks}
        for future in future_to_eval:
            result = future_to_eval[future].result()
            print(f"Evaluation result for {future}: {result}")



# coding: utf-8

"""
Root Signals API

Root Signals JSON API provides a way to access Root Signals using provisioned API token

The version of the OpenAPI document: 1.0.0 (latest)
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
from enum import Enum

from typing_extensions import Self


class DocumentTypeEnum(str, Enum):
    """
    * `csv` - CSV * `pdf` - PDF * `html` - HTML * `docx` - DOCX * `doc` - DOC * `txt` - TXT * `md` - MD * `pptx` - PPTX * `ppt` - PPT * `xlsx` - XLSX * `xls` - XLS * `odt` - ODT * `rtf` - RTF * `xml` - XML * `eml` - EML * `epub` - EPUB
    """

    """
    allowed enum values
    """
    CSV = "csv"
    PDF = "pdf"
    HTML = "html"
    DOCX = "docx"
    DOC = "doc"
    TXT = "txt"
    MD = "md"
    PPTX = "pptx"
    PPT = "ppt"
    XLSX = "xlsx"
    XLS = "xls"
    ODT = "odt"
    RTF = "rtf"
    XML = "xml"
    EML = "eml"
    EPUB = "epub"

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of DocumentTypeEnum from a JSON string"""
        return cls(json.loads(json_str))



from root import RootSignals
from root.skills import EvaluatorDemonstration

client = RootSignals()

# Create an evaluator
network_troubleshooting_evaluator = client.evaluators.create(
    name="Advanced Network Troubleshooting",
    predicate="""Assess the response for technical accuracy and appropriateness in the context of network troubleshooting.
                Is the advice technically sound and relevant to the user's question?
                Does the troubleshooting process effectively address the likely causes of the issue?
                Is the proposed solution valid and safe to implement?

                User question: {{request}}

                Chatbot response: {{response}}
                """,
    intent="To measure the technical accuracy and appropriateness of network troubleshooting responses",
    model="gemini-2.0-flash",
)


# Run first calibration (benchmarking).
test_result = client.evaluators.calibrate_existing(
    evaluator_id=network_troubleshooting_evaluator.id,
    # The test data is a list of lists, where each inner
    # list contains an expected score for a given request and response.
    test_data=[
        [
            "0.1",
            "My internet is not working.",
            "I'm sorry to hear that your internet isn't working. Let's troubleshoot this step by step.",
        ],
        [
            "0.95",
            "My internet is not working.",
            "Okay, let's check some basics. First, can you tell me what operating system your computer is running (Windows, macOS, etc.)? Also, can you check the Ethernet cable connecting your computer to the router to ensure it is securely plugged in at both ends? After confirming these steps, open a command prompt or terminal and run `ping 8.8.8.8`. Let me know the results. If you are using wireless connection, try to move closer to the router and see if that improves the connectivity. If the ping fails consistently, the issue might be with your ISP. If the connection improves closer to the router, consider improving your wireless coverage with a range extender or by repositioning the router.",
        ],
    ],
)

print(test_result[0].result)

# Improve the evaluator with demonstrations,
# penalize the vague "I'm sorry" response by setting an expected score of 0.1
client.evaluators.update(
    evaluator_id=network_troubleshooting_evaluator.id,
    evaluator_demonstrations=[
        EvaluatorDemonstration(
            response="I'm sorry to hear that your internet isn't working. Let's troubleshoot this step by step.",
            request="My internet is not working.",
            score=0.1,
        ),
    ],
)
# Run second calibration
test_result = client.evaluators.calibrate_existing(
    evaluator_id=network_troubleshooting_evaluator.id,
    test_data=[
        [
            "0.1",
            "My internet is not working.",
            "I'm sorry to hear that your internet isn't working. Let's troubleshoot this step by step.",
        ],
    ],
)

# Check the results. See that the vague "I'm sorry" response receives a lower score.
print(test_result[0].result)



from root import RootSignals

# Connect to the Root Signals API
client = RootSignals()

# Run an evaluator with tags to track the execution.
result = client.evaluators.Faithfulness(
    request="What is your return policy for electronics?",
    response="""
    You can return electronics within 30 days of purchase, provided the item is unused and in its original packaging.
    A receipt or proof of purchase is required.",
    """,
    contexts=[
        """Our returns policy for electronics allows returns within 30 days of purchase.
        The item must be unused, in its original packaging, and accompanied by a valid receipt or proof of purchase.
        Refunds will be issued to the original payment method.""",
        """A receipt or proof of purchase is required.""",
    ],
    tags=["production", "v1.23"],
)

# Get the execution log for the evaluator run.
log = client.execution_logs.get(execution_result=result)
print(log)


# And get all the logs with the same tags.
# Also include the evaluation context field in the response which by default is not included.
logs = client.execution_logs.list(tags=["production"], include=["evaluation_context"])
for log in logs:
    print(log.score)
    print(log.evaluation_context.contexts)



import asyncio

from root import RootSignals

# Connect to the Root Signals API
aclient = RootSignals(run_async=True)


async def main():
    response = "This is polite and clear."

    tasks = [
        aclient.evaluators.Politeness(response),
        aclient.evaluators.Clarity(response),
    ]

    response = await asyncio.gather(*tasks)

    for future in response:
        print(f"Evaluation result for {future.evaluator_name}: {future}")



import inspect
from functools import partial

import pytest

from root.execution_logs import ExecutionLogs
from root.judges import AJudge, Judge, Judges
from root.models import Models
from root.objectives import AObjective, Objective, Objectives
from root.skills import AEvaluator, ASkill, Evaluator, Skill, Skills


def _get_method_info(method):
    original_method = method

    while hasattr(method, "__wrapped__"):
        method = method.__wrapped__

    if hasattr(original_method, "__closure__") and original_method.__closure__:
        for cell in original_method.__closure__:
            if isinstance(cell.cell_contents, (partial, type(lambda: None))):
                method = cell.cell_contents
                break

    if isinstance(method, partial):
        method = method.func

    sig = inspect.signature(method)
    docs = inspect.getdoc(method) or ""
    params = [(name, param.default) for name, param in sig.parameters.items() if name not in ("_client", "self")]
    return {
        "params": params,
        "doc": docs,
    }


@pytest.mark.parametrize(
    "class_type",
    [Models, ExecutionLogs, Objectives, Skills, Judges],
)
def test_sync_async_methods_match(class_type):
    """Test that sync and async versions of methods have matching signatures and docs."""

    ignored_methods = {"from_response", "__init__", "__getattr__"}
    instance = class_type(client_context=lambda: None)

    # Get all methods that start with 'a' and their sync counterparts
    async_methods = [
        method
        for method in dir(instance)
        if method.startswith("a") and method[1:] in dir(instance) and method[1:] not in ignored_methods
    ]

    for async_method in async_methods:
        sync_method = async_method[1:]

        sync_fn = getattr(instance, sync_method)
        async_fn = getattr(instance, async_method)

        sync_info = _get_method_info(sync_fn)
        async_info = _get_method_info(async_fn)

        assert sync_info["params"] == async_info["params"], (
            # fmt: off
            f"Parameters don't match for {sync_method}/{async_method}:\n"
            f"Sync: {sync_info['params']}\n"
            f"Async: {async_info['params']}"
            # fmt: on
        )

        sync_doc = sync_info["doc"].lower()
        async_doc = async_info["doc"].replace("Asynchronously ", "").lower()
        assert sync_doc
        assert async_doc

        assert sync_doc == async_doc, (
            f"Docstrings don't match for {sync_method}/{async_method}:\nSync: {sync_doc}\nAsync: {async_doc}"
        )


@pytest.mark.parametrize(
    "sync_class,async_class",
    [
        (Objective, AObjective),
        (Skill, ASkill),
        (Evaluator, AEvaluator),
        (Judge, AJudge),
    ],
)
def test_wrapper_classes_sync_async_methods_match(sync_class, async_class):
    """Test that sync and async versions of wrapper classes have matching signatures and docs."""
    included_methods = {"run", "get", "versions", "evaluate"}  # List of methods we want to verify

    sync_methods = {name for name in dir(sync_class) if name in included_methods}

    for method_name in sync_methods:
        async_name = f"a{method_name}"
        assert async_name in dir(async_class), f"Missing async version of {method_name}"

        sync_method = getattr(sync_class, method_name)
        async_method = getattr(async_class, async_name)

        sync_info = _get_method_info(sync_method)
        async_info = _get_method_info(async_method)

        assert sync_info["params"] == async_info["params"], (
            f"Parameters don't match for {method_name}/{async_name}:\n"
            f"Sync: {sync_info['params']}\n"
            f"Async: {async_info['params']}"
        )

        sync_doc = sync_info["doc"].lower()
        async_doc = async_info["doc"].replace("Asynchronously ", "").lower()
        assert sync_doc
        assert async_doc

        assert sync_doc == async_doc, (
            f"Docstrings don't match for {method_name}/{async_name}:\nSync: {sync_doc}\nAsync: {async_doc}"
        )



from root import RootSignals
from root.validators import Validator

client = RootSignals()

cooking_skill = client.skills.create(
    name="Cooking skill with validators",
    prompt="Find me a good recipe for Italian food.",
    validators=[
        Validator(
            evaluator_name="Cooking recipe evaluator",
            prompt="Is the following a cooking recipe: {{response}}",
            threshold=0.1,
        ),
        Validator(
            evaluator_name="Truthfulness",
            threshold=0.5,
        ),
    ],
)
response = cooking_skill.run()

# Check if the recipe was about cooking
print(response.validation)



from typing import Dict, Literal, Optional, Union

from pydantic import BaseModel, Field

from .generated.openapi_aclient.models.data_loader_type_enum import DataLoaderTypeEnum as _ADataLoaderTypeEnum
from .generated.openapi_client.models.data_loader_type_enum import DataLoaderTypeEnum as _DataLoaderTypeEnum

EngineType = Literal["google", "bing"]


class WebPageDataLoader(BaseModel):
    """Data loader which retrieves data from a web page"""

    type: _DataLoaderTypeEnum = Field(default=_DataLoaderTypeEnum.WEBPAGE, frozen=True)
    name: str = Field(..., description="Name assigned to the data loader")
    url: str = Field(..., description="The URL to retrieve the data from")

    def get_parameters(self) -> Dict[str, str]:
        return {"url": self.url}


class AWebPageDataLoader(BaseModel):
    """Data loader which retrieves data from a web page"""

    type: _ADataLoaderTypeEnum = Field(default=_ADataLoaderTypeEnum.WEBPAGE, frozen=True)
    name: str = Field(..., description="Name assigned to the data loader")
    url: str = Field(..., description="The URL to retrieve the data from")

    def get_parameters(self) -> Dict[str, str]:
        return {"url": self.url}


class WebSearchDataLoader(BaseModel):
    """Data loader which retrieves data from web search"""

    type: _DataLoaderTypeEnum = Field(default=_DataLoaderTypeEnum.WEBSEARCH, frozen=True)
    name: str = Field(..., description="Name assigned to the data loader")
    engine: EngineType = Field(..., description="Search engine to use")
    search_terms: str = Field(..., description="What to search for in the search engine")

    def get_parameters(self) -> Dict[str, str]:
        return {"engine": self.engine, "search_terms": self.search_terms}


class AWebSearchDataLoader(BaseModel):
    """Data loader which retrieves data from web search"""

    type: _ADataLoaderTypeEnum = Field(default=_ADataLoaderTypeEnum.WEBSEARCH, frozen=True)
    name: str = Field(..., description="Name assigned to the data loader")
    engine: EngineType = Field(..., description="Search engine to use")
    search_terms: str = Field(..., description="What to search for in the search engine")

    def get_parameters(self) -> Dict[str, str]:
        return {"engine": self.engine, "search_terms": self.search_terms}


class SqlDataLoader(BaseModel):
    """Data loader which retrieves data from SQL"""

    type: _DataLoaderTypeEnum = Field(default=_DataLoaderTypeEnum.SQL, frozen=True)
    name: str = Field(..., description="Name assigned to the data loader")
    connection_string: str = Field(..., description="PostgreSQL connection string")
    query: str = Field(..., description="SQL query to execute")

    def get_parameters(self) -> Dict[str, str]:
        return {"connection_string": self.connection_string, "query": self.query}


class ASqlDataLoader(BaseModel):
    """Data loader which retrieves data from SQL"""

    type: _ADataLoaderTypeEnum = Field(default=_ADataLoaderTypeEnum.SQL, frozen=True)
    name: str = Field(..., description="Name assigned to the data loader")
    connection_string: str = Field(..., description="PostgreSQL connection string")
    query: str = Field(..., description="SQL query to execute")

    def get_parameters(self) -> Dict[str, str]:
        return {"connection_string": self.connection_string, "query": self.query}


class ApiCallDataLoader(BaseModel):
    """Data loader which retrieves data from an URL"""

    type: _DataLoaderTypeEnum = Field(default=_DataLoaderTypeEnum.API_CALL, frozen=True)
    name: str = Field(..., description="Name assigned to the data loader")
    endpoint: str = Field(..., description="URL to connect to (GET request)")
    headers: Optional[Dict[str, str]] = Field(None, description="Optional request headers")

    def get_parameters(self) -> Dict[str, Union[str, Dict[str, str], None]]:
        return {"endpoint": self.endpoint, "headers": self.headers}


class AApiCallDataLoader(BaseModel):
    """Data loader which retrieves data from an URL"""

    type: _ADataLoaderTypeEnum = Field(default=_ADataLoaderTypeEnum.API_CALL, frozen=True)
    name: str = Field(..., description="Name assigned to the data loader")
    endpoint: str = Field(..., description="URL to connect to (GET request)")
    headers: Optional[Dict[str, str]] = Field(None, description="Optional request headers")

    def get_parameters(self) -> Dict[str, Union[str, Dict[str, str], None]]:
        return {"endpoint": self.endpoint, "headers": self.headers}


DataLoader = Union[WebPageDataLoader, WebSearchDataLoader, SqlDataLoader, ApiCallDataLoader]
ADataLoader = Union[AWebPageDataLoader, AWebSearchDataLoader, ASqlDataLoader, AApiCallDataLoader]



from contextlib import AbstractAsyncContextManager, AbstractContextManager
from typing import (
    Any,
    AsyncContextManager,
    AsyncIterator,
    Callable,
    ContextManager,
    Generic,
    Iterator,
    List,
    Optional,
    TypeVar,
    Union,
)

from pydantic import StrictStr
from typing_extensions import TypeAlias

from .generated import openapi_aclient, openapi_client

T = TypeVar("T")


ClientContextCallable: TypeAlias = Union[
    Callable[[], ContextManager[openapi_client.ApiClient]], Callable[[], AsyncContextManager[openapi_aclient.ApiClient]]
]


# This is internal generic class only to handle duck typing of the
# local codes' correctness.
#
# The original classes returned by partial_list may or may not have
# anything to do with this (in practise they are pydantic BaseModel
# subclasses that have no shared superclass unfortunately).
#
# Update November 2024. - This does not work well with our async iterators
# since we end up in a nested async iterator situation. Not DRY.
# Usable and left in one place with test for convenience.s
class _PartialResult(Generic[T]):
    next: Optional[StrictStr] = None
    results: Optional[List[T]] = None


async def aiterate_cursor_list(partial_list: Any, *, limit: int) -> AsyncIterator[T]:
    # TODO: it would be nice to type partial_list correctly.
    cursor: Optional[StrictStr] = None
    while limit > 0:
        result: _PartialResult[T] = await partial_list(page_size=limit, cursor=cursor)
        if not result.results:
            return

        used_results = result.results[:limit]
        yield used_results  # type: ignore[misc]
        limit -= len(used_results)
        if not (cursor := result.next):
            return


def iterate_cursor_list(partial_list: Any, *, limit: int) -> Iterator[T]:
    # TODO: it would be nice to type partial_list correctly.
    cursor: Optional[StrictStr] = None
    while limit > 0:
        result: _PartialResult[T] = partial_list(page_size=limit, cursor=cursor)
        if not result.results:
            return
        used_results = result.results[:limit]
        yield from used_results
        limit -= len(used_results)
        if not (cursor := result.next):
            return


def with_sync_client(func: Callable) -> Callable:
    def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
        context = self.client_context()
        assert isinstance(context, AbstractContextManager), "This method is not available in asynchronous mode"
        with context as client:
            return func(self, *args, _client=client, **kwargs)

    return wrapper


def with_async_client(func: Callable) -> Callable:
    async def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
        context = self.client_context()
        assert isinstance(context, AbstractAsyncContextManager), "This method is not available in synchronous mode"
        async with context as client:
            return await func(self, *args, _client=client, **kwargs)

    return wrapper



# coding: utf-8

"""
Root Signals API

Root Signals JSON API provides a way to access Root Signals using provisioned API token

The version of the OpenAPI document: 1.0.0 (latest)
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Any, ClassVar, Dict, List, Optional, Set, Union

from pydantic import BaseModel, ConfigDict, Field, StrictFloat, StrictInt
from typing_extensions import Self

from root.generated.openapi_aclient.models.evaluator_calibration_result import EvaluatorCalibrationResult


class EvaluatorCalibrationOutput(BaseModel):
    """
    EvaluatorCalibrationOutput
    """  # noqa: E501

    variables: Dict[str, Any]
    model_call_duration: Optional[Union[StrictFloat, StrictInt]] = Field(
        default=None, description="Deprecated, use result.model_call_duration instead."
    )
    row_number: StrictInt
    result: EvaluatorCalibrationResult
    __properties: ClassVar[List[str]] = ["variables", "model_call_duration", "row_number", "result"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of EvaluatorCalibrationOutput from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of result
        if self.result:
            _dict["result"] = self.result.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of EvaluatorCalibrationOutput from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "variables": obj.get("variables"),
                "model_call_duration": obj.get("model_call_duration"),
                "row_number": obj.get("row_number"),
                "result": EvaluatorCalibrationResult.from_dict(obj["result"])
                if obj.get("result") is not None
                else None,
            }
        )
        return _obj



from __future__ import annotations

import os
import re
import textwrap
from contextlib import asynccontextmanager, contextmanager
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    AsyncContextManager,
    AsyncGenerator,
    Callable,
    ContextManager,
    Generator,
    Optional,
    Type,
    Union,
)

from .__about__ import __version__
from .generated import openapi_aclient, openapi_client
from .generated.openapi_aclient.configuration import Configuration as _AConfiguration
from .generated.openapi_client.configuration import Configuration as _Configuration

if TYPE_CHECKING:
    from .datasets import DataSets
    from .execution_logs import ExecutionLogs
    from .judges import Judges
    from .models import Models
    from .objectives import Objectives
    from .skills import Evaluators, Skills


def _get_api_key(*, dot_env: str = ".env") -> str:
    var = "ROOTSIGNALS_API_KEY"
    api_key = os.environ.get(var)
    if api_key is not None:
        return api_key
    if os.path.exists(dot_env):
        for line in open(dot_env):
            m = re.fullmatch(line, rf"^\s*{var}\s*=\s*(\S+)\s*$")
            if m is not None:
                return m.group(1)
    raise ValueError(
        textwrap.dedent("""
    Root Signals API key cannot be found.
    It can be provided in client invocation, using ROOTSIGNALS_API_KEY environment variable or .env file line
    """)
    )


class Beta:
    """Beta API features namespace"""

    def __init__(
        self,
        get_client_context: Union[
            Callable[[], AsyncContextManager[openapi_aclient.ApiClient]],
            Callable[[], ContextManager[openapi_client.ApiClient]],
        ],
    ) -> None:
        self._get_client_context = get_client_context

    @cached_property
    def judges(self) -> Judges:
        """Get Judges API (Beta)"""
        from .judges import Judges

        return Judges(self._get_client_context)


class RootSignals:
    """Root Signals API Python client.

    The API key must be provided via one of the following methods - the code uses the first one that is found:

    1. as an argument to RootSignals constructor,
    2. environment variable `ROOTSIGNALS_API_KEY`, or
    3. .env file containing `ROOTSIGNALS_API_KEY=`

    Args:
        api_key: Root Signals API Key (if not provided from environment)
        run_async: Whether to run the API client asynchronously
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        *,
        run_async: bool = False,
        _api_client: Union[Optional[openapi_aclient.ApiClient], Optional[openapi_client.ApiClient]] = None,
        base_url: Optional[str] = None,
    ):
        self.run_async = run_async
        if api_key is None:
            api_key = _get_api_key()
        if base_url is None:
            base_url = os.environ.get("ROOTSIGNALS_API_URL", "https://api.app.rootsignals.ai")
        self.base_url = base_url
        self.api_key = api_key
        self._api_client_arg = _api_client

    @cached_property
    def get_client_context(
        self,
    ) -> Union[
        Callable[[], AsyncContextManager[openapi_aclient.ApiClient]],
        Callable[[], ContextManager[openapi_client.ApiClient]],
    ]:
        if self._api_client_arg is not None:
            if isinstance(self._api_client_arg, openapi_aclient.ApiClient):

                @asynccontextmanager
                async def async_client_context() -> AsyncGenerator[openapi_aclient.ApiClient, None]:
                    assert isinstance(self._api_client_arg, openapi_aclient.ApiClient)
                    yield self._api_client_arg

                return async_client_context
            else:

                @contextmanager
                def sync_client_context() -> Generator[openapi_client.ApiClient, None, None]:
                    assert isinstance(self._api_client_arg, openapi_client.ApiClient)
                    yield self._api_client_arg

                return sync_client_context

        if self.run_async:
            return self._configure_client_context(openapi_aclient.ApiClient, _AConfiguration)
        return self._configure_client_context(openapi_client.ApiClient, _Configuration)

    def _configure_client_context(
        self,
        client_cls: Union[Type[openapi_client.ApiClient], Type[openapi_aclient.ApiClient]],
        config_cls: Union[Type[_Configuration], Type[_AConfiguration]],
    ) -> Union[
        Callable[[], ContextManager[openapi_client.ApiClient]],
        Callable[[], AsyncContextManager[openapi_aclient.ApiClient]],
    ]:
        config = config_cls(host=self.base_url)
        config.api_key["publicApiKey"] = f"Api-Key {self.api_key}"

        if issubclass(client_cls, openapi_aclient.ApiClient):

            @asynccontextmanager
            async def async_client_context() -> AsyncGenerator[openapi_aclient.ApiClient, None]:
                async with client_cls(config) as client:
                    client.user_agent = f"rs-python-sdk/{__version__}"
                    yield client

            return async_client_context
        else:

            @contextmanager
            def sync_client_context() -> Generator[openapi_client.ApiClient, None, None]:
                with client_cls(config) as client:
                    client.user_agent = f"rs-python-sdk/{__version__}"
                    yield client

            return sync_client_context

    @cached_property
    def datasets(self) -> DataSets:
        """Get DataSets API"""
        from .datasets import DataSets

        return DataSets(self.get_client_context, self.base_url, self.api_key)

    @cached_property
    def evaluators(self) -> Evaluators:
        """Get Evaluators API"""
        from .skills import Evaluators

        return Evaluators(self.get_client_context)

    @cached_property
    def execution_logs(self) -> ExecutionLogs:
        """Get Execution Logs API"""
        from .execution_logs import ExecutionLogs

        return ExecutionLogs(self.get_client_context)

    @cached_property
    def models(self) -> Models:
        """Get Models API"""
        from .models import Models

        return Models(self.get_client_context)

    @cached_property
    def objectives(self) -> Objectives:
        """Get Objectives API"""
        from .objectives import Objectives

        return Objectives(self.get_client_context)

    @cached_property
    def skills(self) -> Skills:
        """Get Skills API"""
        from .skills import Skills

        return Skills(self.get_client_context)

    @cached_property
    def beta(self) -> Beta:
        """Get Beta API features"""
        return Beta(self.get_client_context)



import pytest

from root.client import RootSignals


@pytest.mark.asyncio
async def test_throws_exception_when_sync_methods_used_with_async_client__fails():
    with pytest.raises(Exception) as e:
        aclient = RootSignals(api_key="fake", run_async=True)

        aclient.evaluators.get_by_name("Whoops, this is a sync method")
    assert str(e.value) == "This method is not available in asynchronous mode"



from __future__ import annotations

import asyncio
import math
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed
from contextlib import AbstractAsyncContextManager, AbstractContextManager
from enum import Enum
from functools import partial
from typing import TYPE_CHECKING, AsyncIterator, Dict, Iterator, List, Literal, Optional, Union, cast

from pydantic import BaseModel, StrictStr

from .data_loader import ADataLoader, DataLoader
from .generated.openapi_aclient import ApiClient as AApiClient
from .generated.openapi_aclient.api.v1_api import V1Api as AObjectivesApi
from .generated.openapi_aclient.api.v1_api import V1Api as ASkillsApi
from .generated.openapi_aclient.models import (
    EvaluatorDemonstrationsRequest as AEvaluatorDemonstrationsRequest,
)
from .generated.openapi_aclient.models import (
    EvaluatorExecutionFunctionsRequest as AEvaluatorExecutionFunctionsRequest,
)
from .generated.openapi_aclient.models import (
    EvaluatorExecutionRequest as AEvaluatorExecutionRequest,
)
from .generated.openapi_aclient.models import (
    EvaluatorExecutionResult as AEvaluatorExecutionResult,
)
from .generated.openapi_aclient.models import (
    ModelParamsRequest as AModelParamsRequest,
)
from .generated.openapi_aclient.models.data_loader_request import DataLoaderRequest as ADataLoaderRequest
from .generated.openapi_aclient.models.evaluator_calibration_output import (
    EvaluatorCalibrationOutput as AEvaluatorCalibrationOutput,
)
from .generated.openapi_aclient.models.input_variable_request import InputVariableRequest as AInputVariableRequest
from .generated.openapi_aclient.models.objective_request import ObjectiveRequest as AObjectiveRequest
from .generated.openapi_aclient.models.paginated_skill_list import PaginatedSkillList as APaginatedSkillList
from .generated.openapi_aclient.models.paginated_skill_list_output_list import (
    PaginatedSkillListOutputList as APaginatedSkillListOutputList,
)
from .generated.openapi_aclient.models.patched_skill_request import PatchedSkillRequest as APatchedSkillRequest
from .generated.openapi_aclient.models.reference_variable_request import (
    ReferenceVariableRequest as AReferenceVariableRequest,
)
from .generated.openapi_aclient.models.skill import Skill as AOpenAPISkill
from .generated.openapi_aclient.models.skill_execution_request import (
    SkillExecutionRequest as ASkillExecutionRequest,
)
from .generated.openapi_aclient.models.skill_execution_result import SkillExecutionResult as ASkillExecutionResult
from .generated.openapi_aclient.models.skill_list_output import SkillListOutput as ASkillListOutput
from .generated.openapi_aclient.models.skill_request import SkillRequest as ASkillRequest
from .generated.openapi_aclient.models.skill_test_data_request import SkillTestDataRequest as ASkillTestDataRequest
from .generated.openapi_aclient.models.skill_test_input_request import (
    SkillTestInputRequest as ASkillTestInputRequest,
)
from .generated.openapi_aclient.models.skill_test_output import SkillTestOutput as ASkillTestOutput
from .generated.openapi_client.api.v1_api import ApiClient
from .generated.openapi_client.api.v1_api import V1Api as ObjectivesApi
from .generated.openapi_client.api.v1_api import V1Api as SkillsApi
from .generated.openapi_client.models.data_loader_request import DataLoaderRequest
from .generated.openapi_client.models.evaluator_calibration_output import EvaluatorCalibrationOutput
from .generated.openapi_client.models.evaluator_demonstrations_request import (
    EvaluatorDemonstrationsRequest,
)
from .generated.openapi_client.models.evaluator_execution_functions_request import (
    EvaluatorExecutionFunctionsRequest,
)
from .generated.openapi_client.models.evaluator_execution_request import EvaluatorExecutionRequest
from .generated.openapi_client.models.evaluator_execution_result import EvaluatorExecutionResult
from .generated.openapi_client.models.input_variable_request import InputVariableRequest
from .generated.openapi_client.models.model_params_request import ModelParamsRequest
from .generated.openapi_client.models.objective_request import ObjectiveRequest
from .generated.openapi_client.models.paginated_skill_list import PaginatedSkillList
from .generated.openapi_client.models.patched_skill_request import PatchedSkillRequest
from .generated.openapi_client.models.reference_variable_request import ReferenceVariableRequest
from .generated.openapi_client.models.skill import Skill as OpenAPISkill
from .generated.openapi_client.models.skill_execution_request import SkillExecutionRequest
from .generated.openapi_client.models.skill_execution_result import SkillExecutionResult
from .generated.openapi_client.models.skill_list_output import SkillListOutput
from .generated.openapi_client.models.skill_request import SkillRequest
from .generated.openapi_client.models.skill_test_data_request import SkillTestDataRequest
from .generated.openapi_client.models.skill_test_input_request import SkillTestInputRequest
from .generated.openapi_client.models.skill_test_output import SkillTestOutput
from .utils import ClientContextCallable, aiterate_cursor_list, iterate_cursor_list, with_async_client, with_sync_client

if TYPE_CHECKING:
    from .validators import AValidator, Validator


ModelName = Union[
    str,
    Literal[
        "root",  # RS-chosen model
    ],
]


class ModelParams(BaseModel):
    """
    Additional model parameters.

    All fields are made optional in practice.
    """

    temperature: Optional[float] = None


class ReferenceVariable(BaseModel):
    """
    Reference variable definition.

    `name` within prompt gets populated with content from `dataset_id`.
    """

    name: str
    dataset_id: str


class InputVariable(BaseModel):
    """
    Input variable definition.

    `name` within prompt gets populated with the provided variable.
    """

    name: str


class EvaluatorDemonstration(BaseModel):
    """
    Evaluator demonstration

    Demonstrations are used to train an evaluator to adjust its behavior.
    """

    request: Optional[str] = None
    response: str
    score: float
    justification: Optional[str] = None


class ACalibrateBatchParameters:
    def __init__(
        self,
        name: str,
        prompt: str,
        model: "ModelName",
        pii_filter: bool = False,
        reference_variables: Optional[Union[List["ReferenceVariable"], List["AReferenceVariableRequest"]]] = None,
        input_variables: Optional[Union[List["InputVariable"], List["AInputVariableRequest"]]] = None,
        data_loaders: Optional[List["ADataLoader"]] = None,
    ):
        self.name = name
        self.prompt = prompt
        self.model = model
        self.pii_filter = pii_filter
        self.reference_variables = reference_variables
        self.input_variables = input_variables
        self.data_loaders = data_loaders


class ACalibrateBatchResult(BaseModel):
    results: List[AEvaluatorCalibrationOutput]
    rms_errors_model: Dict[str, float]
    mae_errors_model: Dict[str, float]
    rms_errors_prompt: Dict[str, float]
    mae_errors_prompt: Dict[str, float]


class CalibrateBatchParameters:
    def __init__(
        self,
        name: str,
        prompt: str,
        model: "ModelName",
        pii_filter: bool = False,
        reference_variables: Optional[Union[List["ReferenceVariable"], List["ReferenceVariableRequest"]]] = None,
        input_variables: Optional[Union[List["InputVariable"], List["InputVariableRequest"]]] = None,
        data_loaders: Optional[List["DataLoader"]] = None,
    ):
        self.name = name
        self.prompt = prompt
        self.model = model
        self.pii_filter = pii_filter
        self.reference_variables = reference_variables
        self.input_variables = input_variables
        self.data_loaders = data_loaders


class CalibrateBatchResult(BaseModel):
    results: List[EvaluatorCalibrationOutput]
    rms_errors_model: Dict[str, float]
    mae_errors_model: Dict[str, float]
    rms_errors_prompt: Dict[str, float]
    mae_errors_prompt: Dict[str, float]


class Versions:
    """
    Version listing (sub)API

    Note that this should not be directly instantiated.
    """

    def __init__(self, client_context: ClientContextCallable):
        self.client_context = client_context

    @with_sync_client
    def list(self, skill_id: str, *, _client: ApiClient) -> PaginatedSkillList:
        """
        List all versions of a skill.
        """

        api_instance = SkillsApi(_client)
        return api_instance.list_skill_versions(id=skill_id)

    async def alist(self, skill_id: str) -> APaginatedSkillList:
        """
        Asynchronously list all versions of a skill.
        """

        context = self.client_context()
        assert isinstance(context, AbstractAsyncContextManager), "This method is not available in synchronous mode"
        async with context as client:
            api_instance = ASkillsApi(client)
            return await api_instance.list_skill_versions(id=skill_id)


class Skill(OpenAPISkill):
    """
    Wrapper for a single Skill.

    For available attributes, please check the (automatically
    generated) superclass documentation.
    """

    client_context: ClientContextCallable

    @classmethod
    def _wrap(cls, apiobj: OpenAPISkill, client_context: ClientContextCallable) -> "Skill":
        if not isinstance(apiobj, OpenAPISkill):
            raise ValueError(f"Wrong instance in _wrap: {apiobj!r}")
        obj = cast(Skill, apiobj)
        obj.__class__ = cls
        obj.client_context = client_context
        return obj

    @property
    def openai_base_url(self) -> str:
        """
        Get the OpenAI compatibility API URL for the skill.

        Currently only OpenAI chat completions API is supported using
        the base URL.
        """
        context = self.client_context()
        assert isinstance(context, AbstractContextManager), "This method is not available in asynchronous mode"
        with context as client:
            return f"{client.configuration._base_path}/skills/openai/{self.id}"

    @property
    async def aopenai_base_url(self) -> str:
        """
        Get the OpenAI compatibility API URL for the skill.

        Currently only OpenAI chat completions API is supported using
        the base URL.
        """
        context = self.client_context()
        assert isinstance(context, AbstractAsyncContextManager), "This method is not available in synchronous mode"
        async with context as client:
            return f"{client.configuration._base_path}/skills/openai/{self.id}"

    @with_sync_client
    def run(
        self,
        variables: Optional[Dict[str, str]] = None,
        *,
        model_params: Optional[Union[ModelParams, ModelParamsRequest]] = None,
        _client: ApiClient,
    ) -> SkillExecutionResult:
        """
        Run a skill.

        Args:
            variables: Dictionary mapping the prompt template variables to their values. For example, if the prompt is
                "tell me about {{subject}}", then variables={"subject": "history"} would generate
                "tell me about history".
            model_params: Optional model parameters (e.g. temperature).
        """

        api_instance = SkillsApi(_client)
        skill_execution_request = SkillExecutionRequest(
            variables=variables,
            skill_version_id=self.version_id,
            model_params=_to_model_params(model_params),
        )
        return api_instance.v1_skills_execute_create(id=self.id, skill_execution_request=skill_execution_request)


class ASkill(AOpenAPISkill):
    """
    Wrapper for a single Skill.

    For available attributes, please check the (automatically
    generated) superclass documentation.
    """

    client_context: ClientContextCallable

    @classmethod
    async def _awrap(cls, apiobj: AOpenAPISkill, client_context: ClientContextCallable) -> "ASkill":
        if not isinstance(apiobj, AOpenAPISkill):
            raise ValueError(f"Wrong instance in _wrap: {apiobj!r}")
        obj = cast(ASkill, apiobj)
        obj.__class__ = cls
        obj.client_context = client_context
        return obj

    @with_async_client
    async def arun(
        self,
        variables: Optional[Dict[str, str]] = None,
        *,
        model_params: Optional[Union[ModelParams, AModelParamsRequest]] = None,
        _client: AApiClient,
    ) -> ASkillExecutionResult:
        """
        Asynchronously run a skill.

        Args:
            variables: Dictionary mapping the prompt template variables to their values. For example, if the prompt is
                "tell me about {{subject}}", then variables={"subject": "history"} would generate
                "tell me about history".
            model_params: Optional model parameters (e.g. temperature).
        """

        api_instance = ASkillsApi(_client)
        skill_execution_request = ASkillExecutionRequest(
            variables=variables,
            skill_version_id=self.version_id,
            model_params=_ato_model_params(model_params),
        )
        return await api_instance.v1_skills_execute_create(id=self.id, skill_execution_request=skill_execution_request)


class Evaluator(OpenAPISkill):
    """
    Wrapper for a single Evaluator.

    For available attributes, please check the (automatically
    generated) superclass documentation.
    """

    client_context: ClientContextCallable

    @classmethod
    def _wrap(cls, apiobj: OpenAPISkill, client_context: ClientContextCallable) -> "Evaluator":
        if not isinstance(apiobj, OpenAPISkill):
            raise ValueError(f"Wrong instance in _wrap: {apiobj!r}")
        obj = cast(Evaluator, apiobj)
        obj.__class__ = cls
        obj.client_context = client_context
        return obj

    @with_sync_client
    def run(
        self,
        response: Optional[str] = None,
        request: Optional[str] = None,
        contexts: Optional[List[str]] = None,
        functions: Optional[List[EvaluatorExecutionFunctionsRequest]] = None,
        expected_output: Optional[str] = None,
        variables: Optional[dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        *,
        _client: ApiClient,
        _request_timeout: Optional[int] = None,
    ) -> EvaluatorExecutionResult:
        """
        Run the evaluator.

        Args:
          response: LLM output.
          request: The prompt sent to the LLM.
          contexts: Optional documents passed to RAG evaluators
          functions: Optional function definitions to LLM tool call validation
          expected_output: Optional expected output for the evaluator.
          variables: Optional additional variable mappings for the evaluator. For example, if the evaluator
            predicate is "evaluate the output based on {subject}: {output}", then variables={"subject": "clarity"}.
          tags: Optional tags to add to the evaluator execution
        """

        if not response and not request:
            raise ValueError("Either response or request must be provided")

        api_instance = SkillsApi(_client)

        evaluator_execution_request = EvaluatorExecutionRequest(
            skill_version_id=self.version_id,
            request=request,
            response=response,
            contexts=contexts,
            functions=functions,
            expected_output=expected_output,
            variables=variables,
            tags=tags,
        )
        return api_instance.v1_skills_evaluator_execute_create(
            id=self.id,
            evaluator_execution_request=evaluator_execution_request,
            _request_timeout=_request_timeout,
        )


class AEvaluator(AOpenAPISkill):
    """
    Wrapper for a single Evaluator.

    For available attributes, please check the (automatically
    generated) superclass documentation.
    """

    client_context: ClientContextCallable

    @classmethod
    async def _awrap(cls, apiobj: AOpenAPISkill, client_context: ClientContextCallable) -> "AEvaluator":
        if not isinstance(apiobj, AOpenAPISkill):
            raise ValueError(f"Wrong instance in _wrap: {apiobj!r}")
        obj = cast(AEvaluator, apiobj)
        obj.__class__ = cls
        obj.client_context = client_context
        return obj

    @with_async_client
    async def arun(
        self,
        response: Optional[str] = None,
        request: Optional[str] = None,
        contexts: Optional[List[str]] = None,
        functions: Optional[List[AEvaluatorExecutionFunctionsRequest]] = None,
        expected_output: Optional[str] = None,
        variables: Optional[dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        *,
        _client: AApiClient,
        _request_timeout: Optional[int] = None,
    ) -> AEvaluatorExecutionResult:
        """
        Asynchronously run the evaluator.

        Args:
          response: LLM output.
          request: The prompt sent to the LLM.
          contexts: Optional documents passed to RAG evaluators
          functions: Optional function definitions to LLM tool call validation
          expected_output: Optional expected output for the evaluator.
          variables: Optional additional variable mappings for the evaluator. For example, if the evaluator
            predicate is "evaluate the output based on {subject}: {output}", then variables={"subject": "clarity"}.
          tags: Optional tags to add to the evaluator execution
        """

        if not response and not request:
            raise ValueError("Either response or request must be provided")

        api_instance = ASkillsApi(_client)

        evaluator_execution_request = AEvaluatorExecutionRequest(
            skill_version_id=self.version_id,
            request=request,
            response=response,
            contexts=contexts,
            functions=functions,
            expected_output=expected_output,
            variables=variables,
            tags=tags,
        )
        return await api_instance.v1_skills_evaluator_execute_create(
            id=self.id,
            evaluator_execution_request=evaluator_execution_request,
            _request_timeout=_request_timeout,
        )


def _to_data_loaders(data_loaders: Optional[List[DataLoader]]) -> List[DataLoaderRequest]:
    return [
        DataLoaderRequest(name=data_loader.name, type=data_loader.type, parameters=data_loader.get_parameters())
        for data_loader in (data_loaders or [])
    ]


def _to_input_variables(
    input_variables: Optional[Union[List[InputVariable], List[InputVariableRequest]]],
) -> List[InputVariableRequest]:
    def _convert_to_generated_model(entry: Union[InputVariable, InputVariableRequest]) -> InputVariableRequest:
        if not isinstance(entry, InputVariableRequest):
            return InputVariableRequest(name=entry.name)
        return entry

    return [_convert_to_generated_model(entry) for entry in input_variables or {}]


def _to_model_params(model_params: Optional[Union[ModelParams, ModelParamsRequest]]) -> Optional[ModelParamsRequest]:
    if isinstance(model_params, ModelParams):
        return ModelParamsRequest(**model_params.model_dump())
    return model_params


def _to_reference_variables(
    reference_variables: Optional[Union[List[ReferenceVariable], List[ReferenceVariableRequest]]],
) -> List[ReferenceVariableRequest]:
    def _convert_to_generated_model(
        entry: Union[ReferenceVariable, ReferenceVariableRequest],
    ) -> ReferenceVariableRequest:
        if not isinstance(entry, ReferenceVariableRequest):
            return ReferenceVariableRequest(name=entry.name, dataset=entry.dataset_id)
        return entry

    return [_convert_to_generated_model(entry) for entry in reference_variables or {}]


def _to_evaluator_demonstrations(
    input_variables: Optional[Union[List[EvaluatorDemonstration], List[EvaluatorDemonstrationsRequest]]],
) -> List[EvaluatorDemonstrationsRequest]:
    def _convert_dict(
        entry: Union[EvaluatorDemonstration, EvaluatorDemonstrationsRequest],
    ) -> EvaluatorDemonstrationsRequest:
        if not isinstance(entry, EvaluatorDemonstrationsRequest):
            return EvaluatorDemonstrationsRequest(
                score=entry.score,
                request=entry.request,
                response=entry.response,
                justification=entry.justification,
            )
        return entry

    return [_convert_dict(entry) for entry in input_variables or {}]


def _ato_data_loaders(data_loaders: Optional[List[ADataLoader]]) -> List[ADataLoaderRequest]:
    return [
        ADataLoaderRequest(name=data_loader.name, type=data_loader.type, parameters=data_loader.get_parameters())
        for data_loader in (data_loaders or [])
    ]


def _ato_input_variables(
    input_variables: Optional[Union[List[InputVariable], List[AInputVariableRequest]]],
) -> List[AInputVariableRequest]:
    def _convert_to_generated_model(entry: Union[InputVariable, AInputVariableRequest]) -> AInputVariableRequest:
        if not isinstance(entry, AInputVariableRequest):
            return AInputVariableRequest(name=entry.name)
        return entry

    return [_convert_to_generated_model(entry) for entry in input_variables or {}]


def _ato_model_params(model_params: Optional[Union[ModelParams, AModelParamsRequest]]) -> Optional[AModelParamsRequest]:
    if isinstance(model_params, ModelParams):
        return AModelParamsRequest(**model_params.model_dump())
    return model_params


def _ato_reference_variables(
    reference_variables: Optional[Union[List[ReferenceVariable], List[AReferenceVariableRequest]]],
) -> List[AReferenceVariableRequest]:
    def _convert_to_generated_model(
        entry: Union[ReferenceVariable, AReferenceVariableRequest],
    ) -> AReferenceVariableRequest:
        if not isinstance(entry, AReferenceVariableRequest):
            return AReferenceVariableRequest(name=entry.name, dataset=entry.dataset_id)
        return entry

    return [_convert_to_generated_model(entry) for entry in reference_variables or {}]


def _ato_evaluator_demonstrations(
    input_variables: Optional[Union[List[EvaluatorDemonstration], List[AEvaluatorDemonstrationsRequest]]],
) -> List[AEvaluatorDemonstrationsRequest]:
    def _aconvert_dict(
        entry: Union[EvaluatorDemonstration, AEvaluatorDemonstrationsRequest],
    ) -> AEvaluatorDemonstrationsRequest:
        if not isinstance(entry, AEvaluatorDemonstrationsRequest):
            return AEvaluatorDemonstrationsRequest(
                score=entry.score,
                request=entry.request,
                response=entry.response,
                justification=entry.justification,
            )
        return entry

    return [_aconvert_dict(entry) for entry in input_variables or {}]


class PresetEvaluatorRunner:
    client_context: ClientContextCallable

    def __init__(
        self,
        client_context: ClientContextCallable,
        skill_id: str,
        eval_name: str,
        evaluator_version_id: Optional[str] = None,
    ):
        self.client_context = client_context
        self.skill_id = skill_id
        self.evaluator_version_id = evaluator_version_id
        self.__name__ = eval_name

    @with_sync_client
    def __call__(
        self,
        response: Optional[str] = None,
        request: Optional[str] = None,
        contexts: Optional[List[str]] = None,
        functions: Optional[List[EvaluatorExecutionFunctionsRequest]] = None,
        expected_output: Optional[str] = None,
        variables: Optional[dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        *,
        _client: ApiClient,
        _request_timeout: Optional[int] = None,
    ) -> EvaluatorExecutionResult:
        """
        Run the evaluator.

        Args:
            response: LLM output.
            request: The prompt sent to the LLM.
            contexts: Optional documents passed to RAG evaluators
            functions: Optional function definitions to LLM tool call validation
            expected_output: Optional expected output for the evaluator.
            variables: Optional additional variable mappings for the evaluator. For example, if the evaluator
                predicate is "evaluate the output based on {subject}: {output}", then variables={"subject": "clarity"}.
            tags: Optional tags to add to the evaluator execution
        """

        if not response and not request:
            raise ValueError("Either response or request must be provided")

        api_instance = SkillsApi(_client)

        evaluator_execution_request = EvaluatorExecutionRequest(
            skill_version_id=self.evaluator_version_id,
            request=request,
            response=response,
            contexts=contexts,
            functions=functions,
            expected_output=expected_output,
            variables=variables,
            tags=tags,
        )
        return api_instance.v1_skills_evaluator_execute_create(
            id=self.skill_id,
            evaluator_execution_request=evaluator_execution_request,
            _request_timeout=_request_timeout,
        )


class APresetEvaluatorRunner:
    client_context: ClientContextCallable

    def __init__(
        self,
        client_context: ClientContextCallable,
        skill_id: str,
        eval_name: str,
        evaluator_version_id: Optional[str] = None,
    ):
        self.client_context = client_context
        self.skill_id = skill_id
        self.evaluator_version_id = evaluator_version_id
        self.__name__ = eval_name

    @with_async_client
    async def __call__(
        self,
        response: Optional[str] = None,
        request: Optional[str] = None,
        contexts: Optional[List[str]] = None,
        functions: Optional[List[AEvaluatorExecutionFunctionsRequest]] = None,
        expected_output: Optional[str] = None,
        variables: Optional[dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        *,
        _client: AApiClient,
        _request_timeout: Optional[int] = None,
    ) -> AEvaluatorExecutionResult:
        """
        Asynchronously run the evaluator.

        Args:
            response: LLM output.
            request: The prompt sent to the LLM.
            contexts: Optional documents passed to RAG evaluators
            functions: Optional function definitions to LLM tool call validation
            expected_output: Optional expected output for the evaluator.
            variables: Optional additional variable mappings for the evaluator. For example, if the evaluator
                predicate is "evaluate the output based on {subject}: {output}", then variables={"subject": "clarity"}.
            tags: Optional tags to add to the evaluator execution
        """

        if not response and not request:
            raise ValueError("Either response or request must be provided")

        api_instance = ASkillsApi(_client)

        evaluator_execution_request = AEvaluatorExecutionRequest(
            skill_version_id=self.evaluator_version_id,
            request=request,
            response=response,
            contexts=contexts,
            functions=functions,
            expected_output=expected_output,
            variables=variables,
            tags=tags,
        )
        return await api_instance.v1_skills_evaluator_execute_create(
            id=self.skill_id,
            evaluator_execution_request=evaluator_execution_request,
            _request_timeout=_request_timeout,
        )


class Skills:
    """
    Skills API

    Note:

      The construction of the API instance should be handled by
      accesing an attribute of a :class:`root.client.RootSignals` instance.
    """

    def __init__(self, client_context: ClientContextCallable):
        self.client_context = client_context
        self.versions = Versions(client_context)

    def _to_objective_request(
        self, *, intent: Optional[str] = None, validators: Optional[List[Validator]] = None
    ) -> ObjectiveRequest:
        return ObjectiveRequest(
            intent=intent,
            validators=[validator._to_request(self) for validator in validators or []],
        )

    async def _ato_objective_request(
        self, *, intent: Optional[str] = None, avalidators: Optional[List[AValidator]] = None
    ) -> AObjectiveRequest:
        return AObjectiveRequest(
            intent=intent,
            validators=[await avalidator._ato_request(self) for avalidator in avalidators or []],
        )

    def _validate_create_params_sanitize_name(
        self, name: Optional[str], intent: Optional[str], validators: Optional[List], objective_id: Optional[str]
    ) -> str:
        if objective_id is not None:
            if intent:
                raise ValueError("Supplying both objective_id and intent is not supported")
            if validators:
                raise ValueError("Supplying both objective_id and validators is not supported")
        if name is None:
            name = "<unnamed>"
        return name

    @with_sync_client
    def create(
        self,
        prompt: str = "",
        *,
        name: Optional[str] = None,
        intent: Optional[str] = None,
        model: Optional[ModelName] = None,
        _client: ApiClient,
        system_message: str = "",
        fallback_models: Optional[List[ModelName]] = None,
        pii_filter: bool = False,
        validators: Optional[List[Validator]] = None,
        reference_variables: Optional[Union[List[ReferenceVariable], List[ReferenceVariableRequest]]] = None,
        input_variables: Optional[Union[List[InputVariable], List[InputVariableRequest]]] = None,
        is_evaluator: Optional[bool] = None,
        data_loaders: Optional[List[DataLoader]] = None,
        model_params: Optional[Union[ModelParams, ModelParamsRequest]] = None,
        evaluator_demonstrations: Optional[List[EvaluatorDemonstration]] = None,
        objective_id: Optional[str] = None,
        overwrite: bool = False,
        _request_timeout: Optional[int] = None,
    ) -> Skill:
        """
        Create a new skill and return the result.

        Args:
            prompt: The prompt template that is provided to the model
            name: Name of the skill (defaulting to <unnamed>)
            intent: The intent of the skill (defaulting to name); not available if objective_id is set
            model: The model to use (defaults to 'root', which means
                Root Signals default at the time of skill creation)
            system_message: The system instruction to give to the model
                (mainly useful with OpenAI compatibility API)
            fallback_models: The fallback models to use in case the primary model fails
            pii_filter: Whether to use PII filter or not
            validators: An optional list of validators; not available if objective_id is set
            reference_variables: An optional list of reference variables for the skill
            input_variables: An optional list of input variables for the skill
            is_evaluator: Whether this skill is an evaluator skill
            data_loaders: An optional list of data loaders.
            model_params: An optional set of additional parameters to the model (e.g., temperature)
            evaluator_demonstrations: Optional list of demonstrations for evaluator skills
            objective_id: Optional pre-existing objective id to assign to the skill
            overwrite: Whether to overwrite a skill with the same name if it exists
            _request_timeout: Optional timeout for the request in seconds
        """

        name = self._validate_create_params_sanitize_name(name, intent, validators, objective_id)
        api_instance = SkillsApi(_client)
        objective: Optional[ObjectiveRequest] = None
        if objective_id is None:
            if intent is None:
                intent = name
            objective = self._to_objective_request(intent=intent, validators=validators)
            objectives_api_instance = ObjectivesApi(_client)
            objective_id = objectives_api_instance.v1_objectives_create(objective_request=objective).id

        skill_request = SkillRequest(
            name=name,
            objective_id=objective_id,
            system_message=system_message,
            prompt=prompt,
            models=[model for model in [model] + (fallback_models or []) if model is not None],
            pii_filter=pii_filter,
            is_evaluator=is_evaluator,
            reference_variables=_to_reference_variables(reference_variables),
            input_variables=_to_input_variables(input_variables),
            data_loaders=_to_data_loaders(data_loaders),
            model_params=_to_model_params(model_params),
            evaluator_demonstrations=_to_evaluator_demonstrations(evaluator_demonstrations),
            overwrite=overwrite,
        )

        skill = api_instance.v1_skills_create(skill_request=skill_request, _request_timeout=_request_timeout)
        return Skill._wrap(skill, self.client_context)

    @with_async_client
    async def acreate(
        self,
        prompt: str = "",
        *,
        name: Optional[str] = None,
        intent: Optional[str] = None,
        model: Optional[ModelName] = None,
        _client: AApiClient,
        system_message: str = "",
        fallback_models: Optional[List[ModelName]] = None,
        pii_filter: bool = False,
        validators: Optional[List[AValidator]] = None,
        reference_variables: Optional[Union[List[ReferenceVariable], List[AReferenceVariableRequest]]] = None,
        input_variables: Optional[Union[List[InputVariable], List[AInputVariableRequest]]] = None,
        is_evaluator: Optional[bool] = None,
        data_loaders: Optional[List[ADataLoader]] = None,
        model_params: Optional[Union[ModelParams, AModelParamsRequest]] = None,
        evaluator_demonstrations: Optional[List[EvaluatorDemonstration]] = None,
        objective_id: Optional[str] = None,
        overwrite: bool = False,
        _request_timeout: Optional[int] = None,
    ) -> ASkill:
        """
        Asynchronously create a new skill and return the result.

        Args:
            prompt: The prompt template that is provided to the model
            name: Name of the skill (defaulting to <unnamed>)
            intent: The intent of the skill (defaulting to name); not available if objective_id is set
            model: The model to use (defaults to 'root', which means
                Root Signals default at the time of skill creation)
            system_message: The system instruction to give to the model
                (mainly useful with OpenAI compatibility API)
            fallback_models: The fallback models to use in case the primary model fails
            pii_filter: Whether to use PII filter or not
            validators: An optional list of validators; not available if objective_id is set
            reference_variables: An optional list of reference variables for the skill
            input_variables: An optional list of input variables for the skill
            is_evaluator: Whether this skill is an evaluator skill
            data_loaders: An optional list of data loaders.
            model_params: An optional set of additional parameters to the model (e.g., temperature)
            evaluator_demonstrations: Optional list of demonstrations for evaluator skills
            objective_id: Optional pre-existing objective id to assign to the skill
            overwrite: Whether to overwrite a skill with the same name if it exists
            _request_timeout: Optional timeout for the request in seconds
        """

        name = self._validate_create_params_sanitize_name(name, intent, validators, objective_id)

        api_instance = ASkillsApi(_client)
        objective: Optional[AObjectiveRequest] = None
        if objective_id is None:
            if intent is None:
                intent = name
            objective = await self._ato_objective_request(intent=intent, avalidators=validators)
            objectives_api_instance = AObjectivesApi(_client)
            new_objective = await objectives_api_instance.v1_objectives_create(objective_request=objective)
            objective_id = new_objective.id

        skill_request = ASkillRequest(
            name=name,
            objective_id=objective_id,
            system_message=system_message,
            prompt=prompt,
            models=[model for model in [model] + (fallback_models or []) if model is not None],
            pii_filter=pii_filter,
            is_evaluator=is_evaluator,
            reference_variables=_ato_reference_variables(reference_variables),
            input_variables=_ato_input_variables(input_variables),
            data_loaders=_ato_data_loaders(data_loaders),
            model_params=_ato_model_params(model_params),
            evaluator_demonstrations=_ato_evaluator_demonstrations(evaluator_demonstrations),
            overwrite=overwrite,
        )

        skill = await api_instance.v1_skills_create(skill_request=skill_request, _request_timeout=_request_timeout)
        return await ASkill._awrap(skill, self.client_context)

    @with_sync_client
    def update(
        self,
        skill_id: str,
        *,
        change_note: Optional[str] = None,
        _client: ApiClient,
        data_loaders: Optional[List[DataLoader]] = None,
        fallback_models: Optional[List[ModelName]] = None,
        input_variables: Optional[Union[List[InputVariable], List[InputVariableRequest]]] = None,
        model: Optional[ModelName] = None,
        name: Optional[str] = None,
        pii_filter: Optional[bool] = None,
        prompt: Optional[str] = None,
        is_evaluator: Optional[bool] = None,
        reference_variables: Optional[Union[List[ReferenceVariable], List[ReferenceVariableRequest]]] = None,
        model_params: Optional[Union[ModelParams, ModelParamsRequest]] = None,
        evaluator_demonstrations: Optional[List[EvaluatorDemonstration]] = None,
        objective_id: Optional[str] = None,
        _request_timeout: Optional[int] = None,
    ) -> Skill:
        """
        Update existing skill instance and return the result.

        For description of the rest of the arguments, please refer to create
        method.
        """

        api_instance = SkillsApi(_client)
        request = PatchedSkillRequest(
            name=name,
            prompt=prompt,
            models=[model] + (fallback_models or []) if model else None,
            pii_filter=pii_filter,
            is_evaluator=is_evaluator,
            reference_variables=_to_reference_variables(reference_variables),
            input_variables=_to_input_variables(input_variables),
            data_loaders=_to_data_loaders(data_loaders),
            change_note=change_note or "",
            model_params=_to_model_params(model_params),
            evaluator_demonstrations=_to_evaluator_demonstrations(evaluator_demonstrations),
            objective_id=objective_id,
        )
        api_response = api_instance.v1_skills_partial_update(
            id=skill_id, patched_skill_request=request, _request_timeout=_request_timeout
        )
        return Skill._wrap(api_response, self.client_context)

    @with_async_client
    async def aupdate(
        self,
        skill_id: str,
        *,
        change_note: Optional[str] = None,
        _client: AApiClient,
        data_loaders: Optional[List[ADataLoader]] = None,
        fallback_models: Optional[List[ModelName]] = None,
        input_variables: Optional[Union[List[InputVariable], List[AInputVariableRequest]]] = None,
        model: Optional[ModelName] = None,
        name: Optional[str] = None,
        pii_filter: Optional[bool] = None,
        prompt: Optional[str] = None,
        is_evaluator: Optional[bool] = None,
        reference_variables: Optional[Union[List[ReferenceVariable], List[AReferenceVariableRequest]]] = None,
        model_params: Optional[Union[ModelParams, AModelParamsRequest]] = None,
        evaluator_demonstrations: Optional[List[EvaluatorDemonstration]] = None,
        objective_id: Optional[str] = None,
        _request_timeout: Optional[int] = None,
    ) -> ASkill:
        """
        Asynchronously update existing skill instance and return the result.

        For description of the rest of the arguments, please refer to create
        method.
        """

        api_instance = ASkillsApi(_client)
        request = APatchedSkillRequest(
            name=name,
            prompt=prompt,
            models=[model] + (fallback_models or []) if model else None,
            pii_filter=pii_filter,
            is_evaluator=is_evaluator,
            reference_variables=_ato_reference_variables(reference_variables),
            input_variables=_ato_input_variables(input_variables),
            data_loaders=_ato_data_loaders(data_loaders),
            change_note=change_note or "",
            model_params=_ato_model_params(model_params),
            evaluator_demonstrations=_ato_evaluator_demonstrations(evaluator_demonstrations),
            objective_id=objective_id,
        )
        api_response = await api_instance.v1_skills_partial_update(
            id=skill_id, patched_skill_request=request, _request_timeout=_request_timeout
        )
        return await ASkill._awrap(api_response, self.client_context)

    @with_sync_client
    def get(
        self,
        skill_id: str,
        _client: ApiClient,
        _request_timeout: Optional[int] = None,
    ) -> Skill:
        """
        Get a Skill instance by ID.
        """

        api_instance = SkillsApi(_client)
        api_response = api_instance.v1_skills_retrieve(id=skill_id, _request_timeout=_request_timeout)
        return Skill._wrap(api_response, self.client_context)

    @with_async_client
    async def aget(
        self,
        skill_id: str,
        _client: AApiClient,
        _request_timeout: Optional[int] = None,
    ) -> ASkill:
        """
        Asynchronously get a Skill instance by ID.
        """

        api_instance = ASkillsApi(_client)
        api_response = await api_instance.v1_skills_retrieve(id=skill_id, _request_timeout=_request_timeout)
        return await ASkill._awrap(api_response, self.client_context)

    @with_sync_client
    def list(
        self,
        search_term: Optional[str] = None,
        *,
        limit: int = 100,
        _client: ApiClient,
        name: Optional[str] = None,
        only_evaluators: bool = False,
        only_root_evaluators: bool = False,
    ) -> Iterator[SkillListOutput]:
        """
        Iterate through the skills.

        Args:
          search_term: Can be used to limit returned skills.
          limit: Number of entries to iterate through at most.
          name: Specific name the returned skills must match.
          only_evaluators: Returns only evaluators.
          only_root_evaluators: Returns only Root Signals defined evaluators.
        """

        api_instance = SkillsApi(_client)
        yield from iterate_cursor_list(
            partial(
                api_instance.v1_skills_list,
                name=name,
                search=search_term,
                is_evaluator=True if only_evaluators else None,
                is_root_evaluator=True if only_root_evaluators else None,
            ),
            limit=limit,
        )

    async def alist(
        self,
        search_term: Optional[str] = None,
        *,
        limit: int = 100,
        name: Optional[str] = None,
        only_evaluators: bool = False,
        only_root_evaluators: bool = False,
    ) -> AsyncIterator[ASkillListOutput]:
        """
        Asynchronously iterate through the skills.

        Args:
          search_term: Can be used to limit returned skills.
          limit: Number of entries to iterate through at most.
          name: Specific name the returned skills must match.
          only_evaluators: Returns only evaluators.
          only_root_evaluators: Returns only Root Signals defined evaluators.
        """

        context = self.client_context()
        assert isinstance(context, AbstractAsyncContextManager), "This method is not available in synchronous mode"
        async with context as client:
            api_instance = ASkillsApi(client)
            partial_list = partial(
                api_instance.v1_skills_list,
                name=name,
                search=search_term,
                is_evaluator=True if only_evaluators else None,
                is_root_evaluator=True if only_root_evaluators else None,
            )

            cursor: Optional[StrictStr] = None
            while limit > 0:
                result: APaginatedSkillListOutputList = await partial_list(page_size=limit, cursor=cursor)
                if not result.results:
                    return

                used_results = result.results[:limit]
                limit -= len(used_results)
                for used_result in used_results:
                    yield used_result

                if not (cursor := result.next):
                    return

    @with_sync_client
    def test(
        self,
        test_dataset_id: str,
        prompt: str,
        model: ModelName,
        *,
        _client: ApiClient,
        fallback_models: Optional[List[ModelName]] = None,
        pii_filter: bool = False,
        validators: Optional[List[Validator]] = None,
        reference_variables: Optional[Union[List[ReferenceVariable], List[ReferenceVariableRequest]]] = None,
        input_variables: Optional[Union[List[InputVariable], List[InputVariableRequest]]] = None,
        data_loaders: Optional[List[DataLoader]] = None,
        _request_timeout: Optional[int] = None,
    ) -> List[SkillTestOutput]:
        """
        Test a skill definition with a test dataset and return the result.

        For description of the rest of the arguments, please refer to create
        method.
        """

        api_instance = SkillsApi(_client)
        skill_test_request = SkillTestInputRequest(
            test_dataset_id=test_dataset_id,
            prompt=prompt,
            models=[model] + (fallback_models or []),
            pii_filter=pii_filter,
            objective=self._to_objective_request(validators=validators),
            reference_variables=_to_reference_variables(reference_variables),
            input_variables=_to_input_variables(input_variables),
            data_loaders=_to_data_loaders(data_loaders),
        )
        return api_instance.v1_skills_test_create(skill_test_request, _request_timeout=_request_timeout)

    @with_async_client
    async def atest(
        self,
        test_dataset_id: str,
        prompt: str,
        model: ModelName,
        *,
        _client: AApiClient,
        fallback_models: Optional[List[ModelName]] = None,
        pii_filter: bool = False,
        validators: Optional[List[AValidator]] = None,
        reference_variables: Optional[Union[List[ReferenceVariable], List[AReferenceVariableRequest]]] = None,
        input_variables: Optional[Union[List[InputVariable], List[AInputVariableRequest]]] = None,
        data_loaders: Optional[List[ADataLoader]] = None,
        _request_timeout: Optional[int] = None,
    ) -> List[ASkillTestOutput]:
        """
        Asynchronously test a skill definition with a test dataset and return the result.

        For description of the rest of the arguments, please refer to create
        method.
        """

        api_instance = ASkillsApi(_client)
        skill_test_request = ASkillTestInputRequest(
            test_dataset_id=test_dataset_id,
            prompt=prompt,
            models=[model] + (fallback_models or []),
            pii_filter=pii_filter,
            objective=await self._ato_objective_request(avalidators=validators),
            reference_variables=_ato_reference_variables(reference_variables),
            input_variables=_ato_input_variables(input_variables),
            data_loaders=_ato_data_loaders(data_loaders),
        )
        return await api_instance.v1_skills_test_create(skill_test_request, _request_timeout=_request_timeout)

    @with_sync_client
    def test_existing(
        self,
        skill_id: str,
        *,
        test_dataset_id: Optional[str] = None,
        test_data: Optional[List[List[str]]] = None,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> List[SkillTestOutput]:
        """
        Test an existing skill.

        Note that only one of the test_data and test_data_set must be provided.

        Args:
          test_data: Ephemeral data to be used to test the skill.
          test_dataset_id: ID of the dataset to be used to test the skill.
        """

        if not test_dataset_id and not test_data:
            raise ValueError("Either test_dataset_id or test_data must be provided")
        if test_dataset_id and test_data:
            raise ValueError("Only one of test_dataset_id or test_data must be provided")
        api_instance = SkillsApi(_client)
        skill_test_request = SkillTestDataRequest(
            test_dataset_id=test_dataset_id,
            test_data=test_data,
        )
        return api_instance.v1_skills_test_create2(skill_id, skill_test_request, _request_timeout=_request_timeout)

    @with_async_client
    async def atest_existing(
        self,
        skill_id: str,
        *,
        test_dataset_id: Optional[str] = None,
        test_data: Optional[List[List[str]]] = None,
        _request_timeout: Optional[int] = None,
        _client: AApiClient,
    ) -> List[ASkillTestOutput]:
        """
        Asynchronously test an existing skill.

        Note that only one of the test_data and test_data_set must be provided.

        Args:
          test_data: Ephemeral data to be used to test the skill.
          test_dataset_id: ID of the dataset to be used to test the skill.
        """

        if not test_dataset_id and not test_data:
            raise ValueError("Either test_dataset_id or test_data must be provided")
        if test_dataset_id and test_data:
            raise ValueError("Only one of test_dataset_id or test_data must be provided")
        api_instance = ASkillsApi(_client)
        skill_test_request = ASkillTestDataRequest(
            test_dataset_id=test_dataset_id,
            test_data=test_data,
        )
        return await api_instance.v1_skills_test_create2(
            skill_id, skill_test_request, _request_timeout=_request_timeout
        )

    @with_sync_client
    def delete(self, skill_id: str, *, _client: ApiClient) -> None:
        """
        Delete the skill.
        """

        api_instance = SkillsApi(_client)
        return api_instance.v1_skills_destroy(id=skill_id)

    @with_async_client
    async def adelete(self, skill_id: str, *, _client: AApiClient) -> None:
        """
        Asynchronously delete the skill.
        """

        api_instance = ASkillsApi(_client)
        return await api_instance.v1_skills_destroy(id=skill_id)

    @with_sync_client
    def run(
        self,
        skill_id: str,
        variables: Optional[Dict[str, str]] = None,
        *,
        model_params: Optional[Union[ModelParams, ModelParamsRequest]] = None,
        skill_version_id: Optional[str] = None,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> SkillExecutionResult:
        """
        Run a skill.

        Args:
            variables: Dictionary mapping the prompt template variables to their values. For example, if the prompt is
                "tell me about {{subject}}", then variables={"subject": "history"} would generate
                "tell me about history".
            model_params: Optional model parameters to override the skill's default parameters
            skill_version_id: Optional version ID of the skill to run. Defaults to the latest version.
            _request_timeout: Optional timeout for the request in seconds
        """

        api_instance = SkillsApi(_client)
        skill_execution_request = SkillExecutionRequest(
            variables=variables,
            skill_version_id=skill_version_id,
            model_params=_to_model_params(model_params),
        )
        return api_instance.v1_skills_execute_create(
            id=skill_id,
            skill_execution_request=skill_execution_request,
            _request_timeout=_request_timeout,
        )

    @with_async_client
    async def arun(
        self,
        skill_id: str,
        variables: Optional[Dict[str, str]] = None,
        *,
        model_params: Optional[Union[ModelParams, AModelParamsRequest]] = None,
        skill_version_id: Optional[str] = None,
        _request_timeout: Optional[int] = None,
        _client: AApiClient,
    ) -> ASkillExecutionResult:
        """
        Asynchronously run a skill.

        Args:
            variables: Dictionary mapping the prompt template variables to their values. For example, if the prompt is
                "tell me about {{subject}}", then variables={"subject": "history"} would generate
                "tell me about history".
            model_params: Optional model parameters to override the skill's default parameters
            skill_version_id: Optional version ID of the skill to run. Defaults to the latest version.
            _request_timeout: Optional timeout for the request in seconds
        """

        api_instance = ASkillsApi(_client)
        skill_execution_request = ASkillExecutionRequest(
            variables=variables,
            skill_version_id=skill_version_id,
            model_params=_ato_model_params(model_params),
        )

        return await api_instance.v1_skills_execute_create(
            id=skill_id,
            skill_execution_request=skill_execution_request,
            _request_timeout=_request_timeout,
        )


class Evaluators:
    """Evaluators (sub) API

    Note:

      The construction of the API instance should be handled by
      accesing an attribute of a :class:`root.client.RootSignals` instance.
    """

    class Eval(Enum):
        # TODO: These eval names should be retrieved automatically from the API or a shared config file
        Faithfulness = "901794f9-634c-4852-9e41-7c558f1ff1ab"
        Relevance = "bd789257-f458-4e9e-8ce9-fa6e86dc3fb9"
        Clarity = "9976d9f3-7265-4732-b518-d61c2642b14e"
        Non_toxicity = "e296e374-7539-4eb2-a74a-47847dd26fb8"
        Helpfulness = "88bc92d5-bebf-45e4-9cd1-dfa33309c320"
        Politeness = "2856903a-e48c-4548-b3fe-520fd88c4f25"
        Formality = "8ab6cf1a-42b5-4a23-a15c-21372816483d"
        Harmlessness = "379fee0a-4fd1-4942-833b-7d78d78b334d"
        Confidentiality = "2eaa0a02-47a9-48f7-9b47-66ad257f93eb"
        Persuasiveness = "85bb6a74-f5dd-4130-8dcc-cffdf72327cc"
        JSON_Empty_Values_Ratio = "03829088-1799-438e-ae30-1db60832e52d"
        JSON_Property_Name_Accuracy = "740923aa-8ffd-49cc-a95d-14f831243b25"
        JSON_Property_Type_Accuracy = "eabc6924-1fec-4e96-82ce-c03bf415c885"
        JSON_Property_Completeness = "e5de37f7-d20c-420f-8072-f41dce96ecfc"
        JSON_Content_Accuracy = "b6a9aeff-c888-46d7-9e9c-7cf8cb461762"
        Context_Recall = "8bb60975-5062-4367-9fc6-a920044cba56"
        Answer_Correctness = "d4487568-4243-4da8-9c76-adbaf762dbe0"
        Answer_Semantic_Similarity = "ff350bce-4b07-4af7-9640-803c9d3c2ff9"
        Sentiment_recognition = "e3782c1e-eaf4-4b2d-8d26-53db2160f1fd"
        Safety_for_Children = "39a8b5ba-de77-4726-a6b0-621d40b3cdf5"
        Precision = "767bdd49-5f8c-48ca-8324-dfd6be7f8a79"
        Originality = "e72cb54f-548a-44f9-a6ca-4e14e5ade7f7"
        Engagingness = "64729487-d4a8-42d8-bd9e-72fd8390c134"
        Conciseness = "be828d33-158a-4e92-a2eb-f4d96c13f956"
        Coherence = "e599886c-c338-458f-91b3-5d7eba452618"
        Quality_of_Writing_Professional = "059affa9-2d1c-48de-8e97-f81dd3fc3cbe"
        Quality_of_Writing_Creative = "060abfb6-57c9-43b5-9a6d-8a1a9bb853b8"
        Truthfulness = "053df10f-b0c7-400b-892e-46ce3aa1e430"
        Context_Precision = "9d1e9a25-7e76-4771-b1e3-40825d7918c5"
        Answer_Relevance = "0907d422-e94f-4c9c-a63d-ec0eefd8a903"

    def __init__(self, client_context: ClientContextCallable):
        self.client_context = client_context
        self.versions = Versions(client_context)

    EvaluatorName = Literal[
        "Faithfulness",
        "Relevance",
        "Clarity",
        "Non_toxicity",
        "Helpfulness",
        "Politeness",
        "Formality",
        "Harmlessness",
        "Confidentiality",
        "Persuasiveness",
        "JSON_Empty_Values_Ratio",
        "JSON_Property_Name_Accuracy",
        "JSON_Property_Type_Accuracy",
        "JSON_Property_Completeness",
        "JSON_Content_Accuracy",
        "Context_Recall",
        "Answer_Correctness",
        "Answer_Semantic_Similarity",
        "Sentiment_recognition",
        "Safety_for_Children",
        "Precision",
        "Originality",
        "Engagingness",
        "Conciseness",
        "Coherence",
        "Quality_of_Writing_Professional",
        "Quality_of_Writing_Creative",
        "Truthfulness",
        "Context_Precision",
        "Answer_Relevance",
    ]

    def __getattr__(self, name: Union[EvaluatorName, str]) -> Union["PresetEvaluatorRunner", "APresetEvaluatorRunner"]:
        if name in self.Eval.__members__:
            context = self.client_context()
            if isinstance(context, AbstractContextManager):
                return PresetEvaluatorRunner(self.client_context, self.Eval.__members__[name].value, name)
            else:
                return APresetEvaluatorRunner(self.client_context, self.Eval.__members__[name].value, name)
        raise AttributeError(f"{name} is not a valid attribute")

    @with_sync_client
    def run(
        self,
        evaluator_id: str,
        *,
        request: Optional[str] = None,
        response: Optional[str] = None,
        contexts: Optional[List[str]] = None,
        functions: Optional[List[EvaluatorExecutionFunctionsRequest]] = None,
        expected_output: Optional[str] = None,
        evaluator_version_id: Optional[str] = None,
        variables: Optional[dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> EvaluatorExecutionResult:
        """
        Run the evaluator.

        Args:
            evaluator_id: The ID of the evaluator to run.
            request: The prompt sent to the LLM.
            response: LLM output.
            contexts: Optional documents passed to RAG evaluators.
            functions: Optional function definitions to LLM tool call validation.
            expected_output: Optional expected output for the evaluator.
            evaluator_version_id: Version ID of the evaluator to run. If omitted, the latest version is used.
            variables: Optional additional variable mappings for the evaluator. For example, if the evaluator
                predicate is "evaluate the output based on {subject}: {output}", then variables={"subject": "clarity"}.
            tags: Optional tags to add to the evaluator execution
            _request_timeout: Optional timeout for the request.
        """

        if not response and not request:
            raise ValueError("Either response or request must be provided")

        api_instance = SkillsApi(_client)
        evaluator_execution_request = EvaluatorExecutionRequest(
            skill_version_id=evaluator_version_id,
            request=request,
            response=response,
            contexts=contexts,
            functions=functions,
            expected_output=expected_output,
            variables=variables,
            tags=tags,
        )
        return api_instance.v1_skills_evaluator_execute_create(
            id=evaluator_id,
            evaluator_execution_request=evaluator_execution_request,
            _request_timeout=_request_timeout,
        )

    @with_async_client
    async def arun(
        self,
        evaluator_id: str,
        *,
        request: Optional[str] = None,
        response: Optional[str] = None,
        contexts: Optional[List[str]] = None,
        functions: Optional[List[AEvaluatorExecutionFunctionsRequest]] = None,
        expected_output: Optional[str] = None,
        evaluator_version_id: Optional[str] = None,
        variables: Optional[dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        _request_timeout: Optional[int] = None,
        _client: AApiClient,
    ) -> AEvaluatorExecutionResult:
        """
        Asynchronously run the evaluator.

        Args:
            evaluator_id: The ID of the evaluator to run.
            request: The prompt sent to the LLM.
            response: LLM output.
            contexts: Optional documents passed to RAG evaluators.
            functions: Optional function definitions to LLM tool call validation.
            expected_output: Optional expected output for the evaluator.
            evaluator_version_id: Version ID of the evaluator to run. If omitted, the latest version is used.
            variables: Optional additional variable mappings for the evaluator. For example, if the evaluator
                predicate is "evaluate the output based on {subject}: {output}", then variables={"subject": "clarity"}.
            tags: Optional tags to add to the evaluator execution
            _request_timeout: Optional timeout for the request.
        """

        if not response and not request:
            raise ValueError("Either response or request must be provided")

        api_instance = ASkillsApi(_client)
        evaluator_execution_request = AEvaluatorExecutionRequest(
            skill_version_id=evaluator_version_id,
            request=request,
            response=response,
            contexts=contexts,
            functions=functions,
            expected_output=expected_output,
            variables=variables,
            tags=tags,
        )
        return await api_instance.v1_skills_evaluator_execute_create(
            id=evaluator_id,
            evaluator_execution_request=evaluator_execution_request,
            _request_timeout=_request_timeout,
        )

    @with_sync_client
    def calibrate_existing(
        self,
        evaluator_id: str,
        *,
        test_dataset_id: Optional[str] = None,
        test_data: Optional[List[List[str]]] = None,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> List[EvaluatorCalibrationOutput]:
        """
        Run calibration set on an existing evaluator.
        """

        if not test_dataset_id and not test_data:
            raise ValueError("Either test_dataset_id or test_data must be provided")
        if test_dataset_id and test_data:
            raise ValueError("Only one of test_dataset_id or test_data must be provided")
        api_instance = SkillsApi(_client)
        skill_test_request = SkillTestDataRequest(
            test_dataset_id=test_dataset_id,
            test_data=test_data,
        )
        return api_instance.v1_skills_calibrate_create2(
            evaluator_id, skill_test_request, _request_timeout=_request_timeout
        )

    @with_async_client
    async def acalibrate_existing(
        self,
        evaluator_id: str,
        *,
        test_dataset_id: Optional[str] = None,
        test_data: Optional[List[List[str]]] = None,
        _request_timeout: Optional[int] = None,
        _client: AApiClient,
    ) -> List[AEvaluatorCalibrationOutput]:
        """
        Asynchronously run calibration set on an existing evaluator.
        """

        if not test_dataset_id and not test_data:
            raise ValueError("Either test_dataset_id or test_data must be provided")
        if test_dataset_id and test_data:
            raise ValueError("Only one of test_dataset_id or test_data must be provided")
        api_instance = ASkillsApi(_client)
        skill_test_request = ASkillTestDataRequest(
            test_dataset_id=test_dataset_id,
            test_data=test_data,
        )
        return await api_instance.v1_skills_calibrate_create2(
            evaluator_id, skill_test_request, _request_timeout=_request_timeout
        )

    @with_sync_client
    def calibrate(
        self,
        *,
        name: str,
        test_dataset_id: Optional[str] = None,
        test_data: Optional[List[List[str]]] = None,
        prompt: str,
        model: ModelName,
        pii_filter: bool = False,
        reference_variables: Optional[Union[List[ReferenceVariable], List[ReferenceVariableRequest]]] = None,
        input_variables: Optional[Union[List[InputVariable], List[InputVariableRequest]]] = None,
        data_loaders: Optional[List[DataLoader]] = None,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> List[EvaluatorCalibrationOutput]:
        """
        Run calibration set for an evaluator definition.
        See the create evaluator method for more details on the parameters.
        """

        if not test_dataset_id and not test_data:
            raise ValueError("Either test_dataset_id or test_data must be provided")
        if test_dataset_id and test_data:
            raise ValueError("Only one of test_dataset_id or test_data must be provided")
        api_instance = SkillsApi(_client)
        skill_test_request = SkillTestInputRequest(
            name=name,
            test_dataset_id=test_dataset_id,
            test_data=test_data,
            prompt=prompt,
            models=[model],
            is_evaluator=True,
            pii_filter=pii_filter,
            objective=ObjectiveRequest(intent="Calibration"),
            reference_variables=_to_reference_variables(reference_variables),
            input_variables=_to_input_variables(input_variables),
            data_loaders=_to_data_loaders(data_loaders),
        )
        return api_instance.v1_skills_calibrate_create(skill_test_request, _request_timeout=_request_timeout)

    @with_async_client
    async def acalibrate(
        self,
        *,
        name: str,
        test_dataset_id: Optional[str] = None,
        test_data: Optional[List[List[str]]] = None,
        prompt: str,
        model: ModelName,
        pii_filter: bool = False,
        reference_variables: Optional[Union[List[ReferenceVariable], List[AReferenceVariableRequest]]] = None,
        input_variables: Optional[Union[List[InputVariable], List[AInputVariableRequest]]] = None,
        data_loaders: Optional[List[ADataLoader]] = None,
        _request_timeout: Optional[int] = None,
        _client: AApiClient,
    ) -> List[AEvaluatorCalibrationOutput]:
        """
        Asynchronously run calibration set for an evaluator definition.
        See the create evaluator method for more details on the parameters.
        """

        if not test_dataset_id and not test_data:
            raise ValueError("Either test_dataset_id or test_data must be provided")
        if test_dataset_id and test_data:
            raise ValueError("Only one of test_dataset_id or test_data must be provided")
        api_instance = ASkillsApi(_client)
        skill_test_request = ASkillTestInputRequest(
            name=name,
            test_dataset_id=test_dataset_id,
            test_data=test_data,
            prompt=prompt,
            models=[model],
            is_evaluator=True,
            pii_filter=pii_filter,
            objective=AObjectiveRequest(intent="Calibration"),
            reference_variables=_ato_reference_variables(reference_variables),
            input_variables=_ato_input_variables(input_variables),
            data_loaders=_ato_data_loaders(data_loaders),
        )
        return await api_instance.v1_skills_calibrate_create(skill_test_request, _request_timeout=_request_timeout)

    def calibrate_batch(
        self,
        *,
        evaluator_definitions: List[CalibrateBatchParameters],
        test_dataset_id: Optional[str] = None,
        test_data: Optional[List[List[str]]] = None,
        parallel_requests: int = 1,
        _request_timeout: Optional[int] = None,
    ) -> CalibrateBatchResult:
        """
        Run calibration for a set of prompts and models

        Args:
             evaluator_definitions: List of evaluator definitions.
             test_dataset_id: ID of the dataset to be used to test the skill.
             test_data: Snapshot of data to be used to test the skill.
             parallel_requests: Number of parallel requests. Uses ThreadPoolExecutor if > 1.
        Returns a model with the results and errors for each model and prompt.
        """

        if test_dataset_id and test_data:
            raise ValueError("Only one of test_dataset_id or test_data must be provided")

        model_errors: Dict[str, Dict[str, float]] = defaultdict(
            lambda: {"sum_squared_errors": 0, "abs_errors": 0, "count": 0}
        )
        prompt_errors: Dict[str, Dict[str, float]] = defaultdict(
            lambda: {"sum_squared_errors": 0, "abs_errors": 0, "count": 0}
        )

        all_results = []

        use_thread_pool = parallel_requests > 1

        def process_results(results: List[EvaluatorCalibrationOutput], param: CalibrateBatchParameters) -> None:
            for result in results:
                score = result.result.score or 0
                expected_score = result.result.expected_score or 0
                squared_error = (score - expected_score) ** 2
                abs_error = abs(score - expected_score)

                # TODO multiple thread race condition
                model_errors[param.model]["sum_squared_errors"] += squared_error
                model_errors[param.model]["abs_errors"] += abs_error
                model_errors[param.model]["count"] += 1

                prompt_errors[param.prompt]["sum_squared_errors"] += squared_error
                prompt_errors[param.prompt]["abs_errors"] += abs_error
                prompt_errors[param.prompt]["count"] += 1

                all_results.append(result)

        if use_thread_pool:
            with ThreadPoolExecutor(max_workers=parallel_requests) as executor:
                futures = {
                    executor.submit(
                        self.calibrate,
                        name=param.name,
                        test_dataset_id=test_dataset_id,
                        test_data=test_data,
                        prompt=param.prompt,
                        model=param.model,
                        pii_filter=param.pii_filter,
                        reference_variables=param.reference_variables,
                        input_variables=param.input_variables,
                        data_loaders=param.data_loaders,
                        _request_timeout=_request_timeout,
                    ): param
                    for param in evaluator_definitions
                }

                for future in as_completed(futures):
                    param = futures[future]
                    try:
                        results = future.result()
                        process_results(results, param)
                    except Exception as exc:
                        raise ValueError(f"Calibration failed for {param.prompt} with model {param.model}") from exc
        else:
            for param in evaluator_definitions:
                try:
                    results = self.calibrate(
                        name=param.name,
                        test_dataset_id=test_dataset_id,
                        test_data=test_data,
                        prompt=param.prompt,
                        model=param.model,
                        pii_filter=param.pii_filter,
                        reference_variables=param.reference_variables,
                        input_variables=param.input_variables,
                        data_loaders=param.data_loaders,
                        _request_timeout=_request_timeout,
                    )
                    process_results(results, param)
                except Exception as exc:
                    raise ValueError(f"Calibration failed for {param.prompt} with model {param.model}") from exc

        rms_errors_model = {
            model: math.sqrt(data["sum_squared_errors"] / data["count"])
            for model, data in model_errors.items()
            if data["count"] > 0
        }

        rms_errors_prompt = {
            prompt: math.sqrt(data["sum_squared_errors"] / data["count"])
            for prompt, data in prompt_errors.items()
            if data["count"] > 0
        }

        mae_errors_model = {
            model: data["abs_errors"] / data["count"] for model, data in model_errors.items() if data["count"] > 0
        }

        mae_errors_prompt = {
            prompt: data["abs_errors"] / data["count"] for prompt, data in prompt_errors.items() if data["count"] > 0
        }

        return CalibrateBatchResult(
            results=all_results,
            rms_errors_model=rms_errors_model,
            rms_errors_prompt=rms_errors_prompt,
            mae_errors_model=mae_errors_model,
            mae_errors_prompt=mae_errors_prompt,
        )

    async def acalibrate_batch(
        self,
        *,
        evaluator_definitions: List[ACalibrateBatchParameters],
        test_dataset_id: Optional[str] = None,
        test_data: Optional[List[List[str]]] = None,
        parallel_requests: int = 1,
        _request_timeout: Optional[int] = None,
    ) -> ACalibrateBatchResult:
        """
        Asynchronously run calibration for a set of prompts and models

        Args:
             evaluator_definitions: List of evaluator definitions.
             test_dataset_id: ID of the dataset to be used to test the skill.
             test_data: Snapshot of data to be used to test the skill.
             parallel_requests: Number of parallel requests.

        Returns a model with the results and errors for each model and prompt.
        """

        if test_dataset_id and test_data:
            raise ValueError("Only one of test_dataset_id or test_data must be provided")

        model_errors: Dict[str, Dict[str, float]] = defaultdict(
            lambda: {"sum_squared_errors": 0, "abs_errors": 0, "count": 0}
        )
        prompt_errors: Dict[str, Dict[str, float]] = defaultdict(
            lambda: {"sum_squared_errors": 0, "abs_errors": 0, "count": 0}
        )

        all_results = []

        async def process_results(results: List[AEvaluatorCalibrationOutput], param: ACalibrateBatchParameters) -> None:
            for result in results:
                score = result.result.score or 0
                expected_score = result.result.expected_score or 0
                squared_error = (score - expected_score) ** 2
                abs_error = abs(score - expected_score)

                model_errors[param.model]["sum_squared_errors"] += squared_error
                model_errors[param.model]["abs_errors"] += abs_error
                model_errors[param.model]["count"] += 1

                prompt_errors[param.prompt]["sum_squared_errors"] += squared_error
                prompt_errors[param.prompt]["abs_errors"] += abs_error
                prompt_errors[param.prompt]["count"] += 1

                all_results.append(result)

        sem = asyncio.Semaphore(parallel_requests)

        async def bounded_calibrate(param: ACalibrateBatchParameters) -> None:
            async with sem:
                try:
                    results = await self.acalibrate(
                        name=param.name,
                        test_dataset_id=test_dataset_id,
                        test_data=test_data,
                        prompt=param.prompt,
                        model=param.model,
                        pii_filter=param.pii_filter,
                        reference_variables=param.reference_variables,
                        input_variables=param.input_variables,
                        data_loaders=param.data_loaders,
                        _request_timeout=_request_timeout,
                    )
                    await process_results(results, param)
                except Exception as exc:
                    raise ValueError(f"Calibration failed for {param.prompt} with model {param.model}") from exc

        await asyncio.gather(*(bounded_calibrate(param) for param in evaluator_definitions))

        rms_errors_model = {
            model: math.sqrt(data["sum_squared_errors"] / data["count"])
            for model, data in model_errors.items()
            if data["count"] > 0
        }

        rms_errors_prompt = {
            prompt: math.sqrt(data["sum_squared_errors"] / data["count"])
            for prompt, data in prompt_errors.items()
            if data["count"] > 0
        }

        mae_errors_model = {
            model: data["abs_errors"] / data["count"] for model, data in model_errors.items() if data["count"] > 0
        }

        mae_errors_prompt = {
            prompt: data["abs_errors"] / data["count"] for prompt, data in prompt_errors.items() if data["count"] > 0
        }

        return ACalibrateBatchResult(
            results=all_results,
            rms_errors_model=rms_errors_model,
            rms_errors_prompt=rms_errors_prompt,
            mae_errors_model=mae_errors_model,
            mae_errors_prompt=mae_errors_prompt,
        )

    @with_sync_client
    def get_by_name(
        self,
        name: str,
        *,
        _client: ApiClient,
    ) -> Evaluator:
        """Get an evaluator instance by name.

        Args:
        name: The evaluator to be fetched. Note this only works for uniquely named evaluators.
        """

        api_instance = SkillsApi(_client)

        evaluator_list: List[SkillListOutput] = list(
            iterate_cursor_list(
                partial(api_instance.v1_skills_list, name=name, is_evaluator=True),
                limit=1,
            )
        )

        if not evaluator_list:
            raise ValueError(f"No evaluator found with name '{name}'")

        evaluator = evaluator_list[0]
        api_response = api_instance.v1_skills_retrieve(id=evaluator.id)

        return Evaluator._wrap(api_response, self.client_context)

    async def aget_by_name(
        self,
        name: str,
    ) -> AEvaluator:
        """Asynchronously get an evaluator instance by name.

        Args:
        name: The evaluator to be fetched. Note this only works for uniquely named evaluators.
        """

        context = self.client_context()

        assert isinstance(context, AbstractAsyncContextManager), "This method is not available in synchronous mode"

        async with context as client:
            api_instance = ASkillsApi(client)

            evaluator_list: List[ASkillListOutput] = []
            async for skill in aiterate_cursor_list(  # type: ignore[var-annotated]
                partial(api_instance.v1_skills_list, name=name, is_evaluator=True),
                limit=1,
            ):
                evaluator_list.extend(skill)

            if not evaluator_list:
                raise ValueError(f"No evaluator found with name '{name}'")

            evaluator = evaluator_list[0]
            api_response = await api_instance.v1_skills_retrieve(id=evaluator.id)

            return await AEvaluator._awrap(api_response, self.client_context)

    def create(
        self,
        predicate: str = "",
        *,
        name: Optional[str] = None,
        intent: Optional[str] = None,
        model: Optional[ModelName] = None,
        fallback_models: Optional[List[ModelName]] = None,
        pii_filter: bool = False,
        reference_variables: Optional[Union[List[ReferenceVariable], List[ReferenceVariableRequest]]] = None,
        input_variables: Optional[Union[List[InputVariable], List[InputVariableRequest]]] = None,
        data_loaders: Optional[List[DataLoader]] = None,
        model_params: Optional[Union[ModelParams, ModelParamsRequest]] = None,
        evaluator_demonstrations: Optional[List[EvaluatorDemonstration]] = None,
        objective_id: Optional[str] = None,
        overwrite: bool = False,
    ) -> Evaluator:
        """Create a new evaluator and return the result

        Args:
          predicate: The question / predicate that is provided to the semantic quantification layer to
          transform it into a final prompt before being passed to the model

          name: Name of the skill (defaulting to <unnamed>)

          objective_id: Optional pre-existing objective id to assign to the evaluator.

          intent: The intent of the skill (defaulting to name); not available if objective_id is set.

          model: The model to use (defaults to 'root', which means
            Root Signals default at the time of skill creation)

          fallback_models: The fallback models to use in case the primary model fails.

          pii_filter: Whether to use PII filter or not.

          reference_variables: An optional list of reference variables for
            the skill.

          input_variables: An optional list of input variables for
            the skill.

          data_loaders: An optional list of data loaders

          model_params: An optional set of additional parameters to the model (e.g., temperature).

          An optional list of evaluator demonstrations to guide
            the evaluator's behavior.

          overwrite: Whether to overwrite a skill with the same name if it exists.
        """

        _eval_skill = Skills(self.client_context).create(
            name=name,
            prompt=predicate,
            model=model,
            intent=intent,
            system_message="",
            fallback_models=fallback_models,
            pii_filter=pii_filter,
            validators=None,
            reference_variables=reference_variables,
            input_variables=input_variables,
            is_evaluator=True,
            data_loaders=data_loaders,
            model_params=model_params,
            objective_id=objective_id,
            evaluator_demonstrations=evaluator_demonstrations,
            overwrite=overwrite,
        )
        return Evaluator._wrap(_eval_skill, self.client_context)

    async def acreate(
        self,
        predicate: str = "",
        *,
        name: Optional[str] = None,
        intent: Optional[str] = None,
        model: Optional[ModelName] = None,
        fallback_models: Optional[List[ModelName]] = None,
        pii_filter: bool = False,
        reference_variables: Optional[Union[List[ReferenceVariable], List[AReferenceVariableRequest]]] = None,
        input_variables: Optional[Union[List[InputVariable], List[AInputVariableRequest]]] = None,
        data_loaders: Optional[List[ADataLoader]] = None,
        model_params: Optional[Union[ModelParams, AModelParamsRequest]] = None,
        evaluator_demonstrations: Optional[List[EvaluatorDemonstration]] = None,
        objective_id: Optional[str] = None,
        overwrite: bool = False,
    ) -> AEvaluator:
        """
        Asynchronously create a new evaluator and return the result

        Args:
          predicate: The question / predicate that is provided to the semantic quantification layer to
          transform it into a final prompt before being passed to the model

          name: Name of the skill (defaulting to <unnamed>)

          objective_id: Optional pre-existing objective id to assign to the evaluator.

          intent: The intent of the skill (defaulting to name); not available if objective_id is set.

          model: The model to use (defaults to 'root', which means
            Root Signals default at the time of skill creation)

          fallback_models: The fallback models to use in case the primary model fails.

          pii_filter: Whether to use PII filter or not.

          reference_variables: An optional list of reference variables for
            the skill.

          input_variables: An optional list of input variables for
            the skill.

          data_loaders: An optional list of data loaders

          model_params: An optional set of additional parameters to the model (e.g., temperature).

          evaluator_demonstrations: An optional list of evaluator demonstrations to guide
            the evaluator's behavior.

          overwrite: Whether to overwrite a skill with the same name if it exists.
        """

        _eval_skill = await Skills(self.client_context).acreate(
            name=name,
            prompt=predicate,
            model=model,
            intent=intent,
            system_message="",
            fallback_models=fallback_models,
            pii_filter=pii_filter,
            validators=None,
            reference_variables=reference_variables,
            input_variables=input_variables,
            is_evaluator=True,
            data_loaders=data_loaders,
            evaluator_demonstrations=evaluator_demonstrations,
            model_params=model_params,
            objective_id=objective_id,
            overwrite=overwrite,
        )
        return await AEvaluator._awrap(_eval_skill, self.client_context)

    def update(
        self,
        evaluator_id: str,
        *,
        change_note: Optional[str] = None,
        data_loaders: Optional[List[DataLoader]] = None,
        fallback_models: Optional[List[ModelName]] = None,
        input_variables: Optional[Union[List[InputVariable], List[InputVariableRequest]]] = None,
        model: Optional[ModelName] = None,
        name: Optional[str] = None,
        pii_filter: Optional[bool] = None,
        predicate: Optional[str] = None,
        reference_variables: Optional[Union[List[ReferenceVariable], List[ReferenceVariableRequest]]] = None,
        model_params: Optional[Union[ModelParams, ModelParamsRequest]] = None,
        evaluator_demonstrations: Optional[List[EvaluatorDemonstration]] = None,
        objective_id: Optional[str] = None,
        _request_timeout: Optional[int] = None,
    ) -> Evaluator:
        """
        Update an evaluator and return the result

        See the create method for more information on the arguments.
        """

        evaluator = Skills(self.client_context).update(
            skill_id=evaluator_id,
            change_note=change_note,
            data_loaders=data_loaders,
            fallback_models=fallback_models,
            input_variables=input_variables,
            model=model,
            name=name,
            pii_filter=pii_filter,
            prompt=predicate,
            reference_variables=reference_variables,
            model_params=model_params,
            evaluator_demonstrations=evaluator_demonstrations,
            objective_id=objective_id,
            _request_timeout=_request_timeout,
        )
        return Evaluator._wrap(evaluator, self.client_context)

    async def aupdate(
        self,
        evaluator_id: str,
        *,
        change_note: Optional[str] = None,
        data_loaders: Optional[List[ADataLoader]] = None,
        fallback_models: Optional[List[ModelName]] = None,
        input_variables: Optional[Union[List[InputVariable], List[AInputVariableRequest]]] = None,
        model: Optional[ModelName] = None,
        name: Optional[str] = None,
        pii_filter: Optional[bool] = None,
        predicate: Optional[str] = None,
        reference_variables: Optional[Union[List[ReferenceVariable], List[AReferenceVariableRequest]]] = None,
        model_params: Optional[Union[ModelParams, AModelParamsRequest]] = None,
        evaluator_demonstrations: Optional[List[EvaluatorDemonstration]] = None,
        objective_id: Optional[str] = None,
        _request_timeout: Optional[int] = None,
    ) -> AEvaluator:
        """
        Asynchronously update an evaluator and return the result

        See the create method for more information on the arguments.
        """

        evaluator = await Skills(self.client_context).aupdate(
            skill_id=evaluator_id,
            change_note=change_note,
            data_loaders=data_loaders,
            fallback_models=fallback_models,
            input_variables=input_variables,
            model=model,
            name=name,
            pii_filter=pii_filter,
            prompt=predicate,
            reference_variables=reference_variables,
            model_params=model_params,
            evaluator_demonstrations=evaluator_demonstrations,
            objective_id=objective_id,
            _request_timeout=_request_timeout,
        )
        return await AEvaluator._awrap(evaluator, self.client_context)

    @with_sync_client
    def get(
        self,
        skill_id: str,
        *,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> Evaluator:
        """
        Get a Evaluator instance by ID.
        """

        api_instance = SkillsApi(_client)
        api_response = api_instance.v1_skills_retrieve(id=skill_id, _request_timeout=_request_timeout)
        if not api_response.is_evaluator:
            raise ValueError(f"Skill with id {skill_id} is not an evaluator")
        return Evaluator._wrap(api_response, self.client_context)

    @with_async_client
    async def aget(
        self,
        skill_id: str,
        *,
        _request_timeout: Optional[int] = None,
        _client: AApiClient,
    ) -> AEvaluator:
        """
        Asynchronously get a Evaluator instance by ID.
        """

        api_instance = ASkillsApi(_client)
        api_response = await api_instance.v1_skills_retrieve(id=skill_id, _request_timeout=_request_timeout)
        if not api_response.is_evaluator:
            raise ValueError(f"Skill with id {skill_id} is not an evaluator")
        return await AEvaluator._awrap(api_response, self.client_context)

    def list(
        self,
        search_term: Optional[str] = None,
        *,
        limit: int = 100,
        name: Optional[str] = None,
        only_root_evaluators: bool = False,
    ) -> Iterator[SkillListOutput]:
        """
        Iterate through the evaluators.

        Args:
          search_term: Can be used to limit returned evaluators.
          limit: Number of entries to iterate through at most.
          name: Specific name the returned evaluators must match.
          only_root_evaluators: Returns only Root Signals defined evaluators.
        """

        return Skills(self.client_context).list(
            search_term=search_term,
            limit=limit,
            name=name,
            only_root_evaluators=only_root_evaluators,
            only_evaluators=True,
        )

    async def alist(
        self,
        search_term: Optional[str] = None,
        *,
        limit: int = 100,
        name: Optional[str] = None,
        only_root_evaluators: bool = False,
    ) -> AsyncIterator[ASkillListOutput]:
        """
        Asynchronously iterate through the evaluators.

        Args:
          search_term: Can be used to limit returned evaluators.
          limit: Number of entries to iterate through at most.
          name: Specific name the returned evaluators must match.
          only_root_evaluators: Returns only Root Signals defined evaluators.
        """

        return Skills(self.client_context).alist(
            search_term=search_term,
            limit=limit,
            name=name,
            only_root_evaluators=only_root_evaluators,
            only_evaluators=True,
        )
